from opendr.perception.activity_recognition.x3d.x3d_learner import X3DLearner
from opendr.perception.activity_recognition.cox3d.cox3d_learner import CoX3DLearner
from opendr.perception.activity_recognition.continual_transformer_encoder.continual_transformer_encoder_learner import (
    CoTransEncLearner,
)

from opendr.perception.activity_recognition.datasets.kinetics import (
    KineticsDataset,
    CLASSES,
)

__all__ = [
    "X3DLearner",
    "CoX3DLearner",
    "CoTransEncLearner",
    "KineticsDataset",
    "CLASSES",
]



from collections import OrderedDict
from typing import Callable

import continual as co
import torch
from continual import PaddingMode
from torch import nn


def CoResBlock(
    dim_in,
    dim_out,
    temp_kernel_size,
    stride,
    trans_func,
    dim_inner,
    num_groups=1,
    stride_1x1=False,
    inplace_relu=True,
    eps=1e-5,
    bn_mmt=0.1,
    dilation=1,
    norm_module=torch.nn.BatchNorm3d,
    block_idx=0,
    drop_connect_rate=0.0,
    temporal_window_size: int = 4,
    temporal_fill: PaddingMode = "zeros",
    se_scope="frame",  # "clip" or "frame"
):
    """
    ResBlock class constructs redisual blocks. More details can be found in:
        Kaiming He, Xiangyu Zhang, Shaoqing Ren, and Jian Sun.
        "Deep residual learning for image recognition."
        https://arxiv.org/abs/1512.03385
    Args:
        dim_in (int): the channel dimensions of the input.
        dim_out (int): the channel dimension of the output.
        temp_kernel_size (int): the temporal kernel sizes of the middle
            convolution in the bottleneck.
        stride (int): the stride of the bottleneck.
        trans_func (string): transform function to be used to construct the
            bottleneck.
        dim_inner (int): the inner dimension of the block.
        num_groups (int): number of groups for the convolution. num_groups=1
            is for standard ResNet like networks, and num_groups>1 is for
            ResNeXt like networks.
        stride_1x1 (bool): if True, apply stride to 1x1 conv, otherwise
            apply stride to the 3x3 conv.
        inplace_relu (bool): calculate the relu on the original input
            without allocating new memory.
        eps (float): epsilon for batch norm.
        bn_mmt (float): momentum for batch norm. Noted that BN momentum in
            PyTorch = 1 - BN momentum in Caffe2.
        dilation (int): size of dilation.
        norm_module (torch.nn.Module): torch.nn.Module for the normalization layer. The
            default is torch.nn.BatchNorm3d.
        drop_connect_rate (float): basic rate at which blocks are dropped,
            linearly increases from input to output blocks.
    """
    branch2 = trans_func(
        dim_in,
        dim_out,
        temp_kernel_size,
        stride,
        dim_inner,
        num_groups,
        stride_1x1=stride_1x1,
        inplace_relu=inplace_relu,
        dilation=dilation,
        norm_module=norm_module,
        block_idx=block_idx,
        temporal_window_size=temporal_window_size,
        temporal_fill=temporal_fill,
        se_scope=se_scope,
    )

    def _is_training(module: nn.Module) -> bool:
        return module.training

    def _drop_connect(x, drop_ratio):
        """Apply dropconnect to x"""
        keep_ratio = 1.0 - drop_ratio
        mask = torch.empty([x.shape[0], 1, 1, 1], dtype=x.dtype, device=x.device)
        mask.bernoulli_(keep_ratio)
        x.div_(keep_ratio)
        x.mul_(mask)
        return x

    if drop_connect_rate > 0:
        drop = [("drop", co.Conditional(_is_training, co.Lambda(_drop_connect)))]
    else:
        drop = []

    main_stream = co.Sequential(
        OrderedDict(
            [
                ("branch2", branch2),
                *drop,
            ]
        )
    )

    if (dim_in == dim_out) and (stride == 1):
        residual_stream = co.Delay(main_stream.delay)
    else:
        residual_stream = co.Sequential(
            OrderedDict(
                [
                    (
                        "branch1",
                        co.Conv3d(
                            dim_in,
                            dim_out,
                            kernel_size=1,
                            stride=(1, stride, stride),
                            padding=0,
                            bias=False,
                            dilation=1,
                        ),
                    ),
                    (
                        "branch1_bn",
                        norm_module(num_features=dim_out, eps=eps, momentum=bn_mmt),
                    ),
                ]
            )
        )

    return co.Sequential(
        co.BroadcastReduce(residual_stream, main_stream, reduce="sum"),
        nn.ReLU(),
    )


def CoResStage(
    dim_in: int,
    dim_out: int,
    stride: int,
    temp_kernel_sizes: int,
    num_blocks: int,
    dim_inner: int,
    num_groups: int,
    num_block_temp_kernel: int,
    dilation: int,
    trans_func: Callable,
    stride_1x1=False,
    inplace_relu=True,
    norm_module=torch.nn.BatchNorm3d,
    drop_connect_rate=0.0,
    temporal_window_size: int = 4,
    temporal_fill: PaddingMode = "zeros",
    se_scope="frame",
    *args,
    **kwargs,
):
    """
    Create a Continual Residual X3D Stage.

    Note: Compared to the original implementation of X3D, we discard the
    obsolete handling of the multiple pathways and the non-local mechanism.

    Args:
        dim_in (int): channel dimensions of the input.
        dim_out (int): channel dimensions of the output.
        temp_kernel_sizes (int): temporal kernel sizes of the
            convolution in the bottleneck.
        stride (int): stride of the bottleneck.
        num_blocks (int): numbers of blocks.
        dim_inner (int): inner channel dimensions of the input.
        num_groups (int): number of groups for the convolution.
            num_groups=1 is for standard ResNet like networks, and
            num_groups>1 is for ResNeXt like networks.
        num_block_temp_kernel (int): extent the temp_kernel_sizes to
            num_block_temp_kernel blocks, then fill temporal kernel size
            of 1 for the rest of the layers.
        dilation (int): size of dilation.
        trans_func (Callable): transformation function to apply on the network.
        norm_module (nn.Module): nn.Module for the normalization layer. The
            default is nn.BatchNorm3d.
        drop_connect_rate (float): basic rate at which blocks are dropped,
            linearly increases from input to output blocks.
    """

    assert num_block_temp_kernel <= num_blocks

    temp_kernel_sizes = (temp_kernel_sizes * num_blocks)[:num_block_temp_kernel] + (
        [1] * (num_blocks - num_block_temp_kernel)
    )

    return co.Sequential(
        OrderedDict(
            [
                (
                    f"pathway0_res{i}",
                    CoResBlock(
                        dim_in=dim_in if i == 0 else dim_out,
                        dim_out=dim_out,
                        temp_kernel_size=temp_kernel_sizes[i],
                        stride=stride if i == 0 else 1,
                        trans_func=trans_func,
                        dim_inner=dim_inner,
                        num_groups=num_groups,
                        stride_1x1=stride_1x1,
                        inplace_relu=inplace_relu,
                        dilation=dilation,
                        norm_module=norm_module,
                        block_idx=i,
                        drop_connect_rate=drop_connect_rate,
                        temporal_window_size=temporal_window_size,
                        temporal_fill=temporal_fill,
                        se_scope=se_scope,
                    ),
                )
                for i in range(num_blocks)
            ]
        )
    )


def c2_msra_fill(module: torch.nn.Module) -> None:
    """
    Initialize `module.weight` using the "MSRAFill" implemented in Caffe2.
    Also initializes `module.bias` to 0.

    Args:
        module (torch.nn.Module): module to initialize.
    """
    # pyre-ignore
    torch.nn.init.kaiming_normal_(module.weight, mode="fan_out", nonlinearity="relu")
    if module.bias is not None:  # pyre-ignore
        torch.nn.init.constant_(module.bias, 0)


def init_weights(model, fc_init_std=0.01, zero_init_final_bn=True):
    """
    Performs ResNet style weight initialization.
    Args:
        fc_init_std (float): the expected standard deviation for fc layer.
        zero_init_final_bn (bool): if True, zero initialize the final bn for
            every bottleneck.
    """
    for m in model.modules():
        if isinstance(m, torch.nn.Conv3d) or isinstance(m, co.Conv3d):
            """
            Follow the initialization method proposed in:
            {He, Kaiming, et al.
            "Delving deep into rectifiers: Surpassing human-level
            performance on imagenet classification."
            arXiv preprint arXiv:1502.01852 (2015)}
            """
            c2_msra_fill(m)
        elif isinstance(m, torch.nn.BatchNorm3d) or isinstance(m, torch.nn.BatchNorm2d):
            if (
                hasattr(m, "transform_final_bn") and
                m.transform_final_bn and
                zero_init_final_bn
            ):
                batchnorm_weight = 0.0
            else:
                batchnorm_weight = 1.0
            if m.weight is not None:
                m.weight.data.fill_(batchnorm_weight)
            if m.bias is not None:
                m.bias.data.zero_()
        if isinstance(m, torch.nn.Linear):
            m.weight.data.normal_(mean=0.0, std=fc_init_std)
            if m.bias is not None:
                m.bias.data.zero_()



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import pytorch_lightning as pl


class _LightningModuleWithCrossEntropy(pl.LightningModule):
    def __init__(self, module):
        pl.LightningModule.__init__(self)
        self.module = module

    def forward(self, *args, **kwargs):
        return self.module.forward(*args, **kwargs)

    def training_step(self, batch, batch_idx):
        x, y = batch
        z = self.module(x)
        loss = torch.nn.functional.cross_entropy(z, y)
        self.log("train/loss", loss)
        self.log("train/acc", _accuracy(z, y))
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        z = self.forward(x)
        loss = torch.nn.functional.cross_entropy(z, y)
        self.log("val/loss", loss)
        self.log("val/acc", _accuracy(z, y))
        return loss

    def test_step(self, batch, batch_idx):
        x, y = batch
        z = self.forward(x)
        loss = torch.nn.functional.cross_entropy(z, y)
        self.log("test/loss", loss)
        self.log("test/acc", _accuracy(z, y))
        return loss


def _accuracy(x, y):
    return torch.sum(x.argmax(dim=1) == y) / len(y)



""" Adapted from: https://github.com/facebookresearch/SlowFast
"""


import torch
import torch.nn as nn


class Swish(nn.Module):
    """Swish activation function: x * sigmoid(x)."""

    def __init__(self):
        super(Swish, self).__init__()

    def forward(self, x):
        return x * torch.sigmoid(x)


class SE(nn.Module):
    """Squeeze-and-Excitation (SE) block w/ Swish: AvgPool, FC, Swish, FC, Sigmoid."""

    def _round_width(self, width, multiplier, min_width=8, divisor=8):
        """
        Round width of filters based on width multiplier
        Args:
            width (int): the channel dimensions of the input.
            multiplier (float): the multiplication factor.
            min_width (int): the minimum width after multiplication.
            divisor (int): the new width should be dividable by divisor.
        """
        if not multiplier:
            return width

        width *= multiplier
        min_width = min_width or divisor
        width_out = max(min_width, int(width + divisor / 2) // divisor * divisor)
        if width_out < 0.9 * width:
            width_out += divisor
        return int(width_out)

    def __init__(self, dim_in, ratio, relu_act=True):
        """
        Args:
            dim_in (int): the channel dimensions of the input.
            ratio (float): the channel reduction ratio for squeeze.
            relu_act (bool): whether to use ReLU activation instead
                of Swish (default).
            divisor (int): the new width should be dividable by divisor.
        """
        super(SE, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool3d((1, 1, 1))
        dim_fc = self._round_width(dim_in, ratio)
        self.fc1 = nn.Conv3d(dim_in, dim_fc, 1, bias=True)
        self.fc1_act = nn.ReLU() if relu_act else Swish()
        self.fc2 = nn.Conv3d(dim_fc, dim_in, 1, bias=True)

        self.fc2_sig = nn.Sigmoid()

    def forward(self, x):
        x_in = x
        for module in self.children():
            x = module(x)
        return x_in * x



""" Adapted from: https://github.com/facebookresearch/SlowFast
"""

import math
import random

import numpy as np
import torch
import torchvision.io as io

from logging import getLogger

logger = getLogger(__name__)


def temporal_sampling(frames, start_idx, end_idx, num_samples):
    """
    Given the start and end frame index, sample num_samples frames between
    the start and end with equal interval.
    Args:
        frames (tensor): a tensor of video frames, dimension is
            `num video frames` x `channel` x `height` x `width`.
        start_idx (int): the index of the start frame.
        end_idx (int): the index of the end frame.
        num_samples (int): number of frames to sample.
    Returns:
        frames (tersor): a tensor of temporal sampled video frames, dimension is
            `num clip frames` x `channel` x `height` x `width`.
    """
    index = torch.linspace(start_idx, end_idx, num_samples)
    index = torch.clamp(index, 0, frames.shape[0] - 1).long()

    # int((end_idx - (frames.shape[0] - 1)) / ((end_idx - start_idx) / num_samples))
    num_padded_frames = len(index) - len(set(index))
    if num_padded_frames > 0:
        logger.warning(
            f"More frames were asked for than are available. Last frame is repeated {num_padded_frames} times as fill."
        )
    frames = torch.index_select(frames, 0, index)
    return frames


def get_start_end_idx(video_size, clip_size, clip_idx, num_clips):
    """
    Sample a clip of size clip_size from a video of size video_size and
    return the indices of the first and last frame of the clip. If clip_idx is
    -1, the clip is randomly sampled, otherwise uniformly split the video to
    num_clips clips, and select the start and end index of clip_idx-th video
    clip.
    Args:
        video_size (int): number of overall frames.
        clip_size (int): size of the clip to sample from the frames.
        clip_idx (int): if clip_idx is -1, perform random jitter sampling. If
            clip_idx is larger than -1, uniformly split the video to num_clips
            clips, and select the start and end index of the clip_idx-th video
            clip.
        num_clips (int): overall number of clips to uniformly sample from the
            given video for testing.
    Returns:
        start_idx (int): the start frame index.
        end_idx (int): the end frame index.
    """
    delta = max(video_size - clip_size, 0)
    if clip_idx == -1:
        # Random temporal sampling.
        start_idx = random.uniform(0, delta)
    else:
        # Uniformly sample the clip with the given index.
        start_idx = delta * clip_idx / num_clips
    end_idx = start_idx + clip_size - 1
    return start_idx, end_idx


def get_start_end_idx_centered(video_size, clip_size, clip_idx, num_clips):

    max_start_idx = max(video_size - clip_size, 0)
    if clip_idx == -1:
        # Random temporal sampling.
        start_idx = random.uniform(0, max_start_idx)
    else:
        segment_size = video_size / num_clips
        segment_center_index = clip_idx * segment_size + segment_size / 2
        start_idx = segment_center_index - clip_size / 2

    # Handle bound for start and end index
    start_idx = max(0, start_idx)
    end_idx = start_idx + clip_size - 1

    return start_idx, end_idx


def pyav_decode_stream(
    container, start_pts, end_pts, stream, stream_name, buffer_size=0
):
    """
    Decode the video with PyAV decoder.
    Args:
        container (container): PyAV container.
        start_pts (int): the starting Presentation TimeStamp to fetch the
            video frames.
        end_pts (int): the ending Presentation TimeStamp of the decoded frames.
        stream (stream): PyAV stream.
        stream_name (dict): a dictionary of streams. For example, {"video": 0}
            means video stream at stream index 0.
        buffer_size (int): number of additional frames to decode beyond end_pts.
    Returns:
        result (list): list of frames decoded.
        max_pts (int): max Presentation TimeStamp of the video sequence.
    """
    # Seeking in the stream is imprecise. Thus, seek to an ealier PTS by a
    # margin pts.
    margin = 1024
    seek_offset = max(start_pts - margin, 0)

    container.seek(seek_offset, any_frame=False, backward=True, stream=stream)
    frames = {}
    buffer_count = 0
    max_pts = 0
    for frame in container.decode(**stream_name):
        max_pts = max(max_pts, frame.pts)
        if frame.pts < start_pts:
            continue
        if frame.pts <= end_pts:
            frames[frame.pts] = frame
        else:
            buffer_count += 1
            frames[frame.pts] = frame
            if buffer_count >= buffer_size:
                break
    result = [frames[pts] for pts in sorted(frames)]
    return result, max_pts


def torchvision_decode(
    video_handle,
    sampling_rate,
    num_frames,
    clip_idx,
    video_meta,
    num_clips=10,
    target_fps=30,
    modalities=("visual",),
    max_spatial_scale=0,
):
    """
    If video_meta is not empty, perform temporal selective decoding to sample a
    clip from the video with TorchVision decoder. If video_meta is empty, decode
    the entire video and update the video_meta.
    Args:
        video_handle (bytes): raw bytes of the video file.
        sampling_rate (int): frame sampling rate (interval between two sampled
            frames).
        num_frames (int): number of frames to sample.
        clip_idx (int): if clip_idx is -1, perform random temporal
            sampling. If clip_idx is larger than -1, uniformly split the
            video to num_clips clips, and select the clip_idx-th video clip.
        video_meta (dict): a dict contains VideoMetaData. Details can be found
            at `pytorch/vision/torchvision/io/_video_opt.py`.
        num_clips (int): overall number of clips to uniformly sample from the
            given video.
        target_fps (int): the input video may has different fps, convert it to
            the target video fps.
        modalities (tuple): tuple of modalities to decode. Currently only
            support `visual`, planning to support `acoustic` soon.
        max_spatial_scale (int): the maximal resolution of the spatial shorter
            edge size during decoding.
    Returns:
        frames (tensor): decoded frames from the video.
        fps (float): the number of frames per second of the video.
        decode_all_video (bool): if True, the entire video was decoded.
    """
    # Convert the bytes to a tensor.
    video_tensor = torch.from_numpy(np.frombuffer(video_handle, dtype=np.uint8))

    decode_all_video = True
    video_start_pts, video_end_pts = 0, -1
    # The video_meta is empty, fetch the meta data from the raw video.
    if len(video_meta) == 0:
        # Tracking the meta info for selective decoding in the future.
        meta = io._probe_video_from_memory(video_tensor)
        # Using the information from video_meta to perform selective decoding.
        video_meta["video_timebase"] = meta.video_timebase
        video_meta["video_numerator"] = meta.video_timebase.numerator
        video_meta["video_denominator"] = meta.video_timebase.denominator
        video_meta["has_video"] = meta.has_video
        video_meta["video_duration"] = meta.video_duration
        video_meta["video_fps"] = meta.video_fps
        video_meta["audio_timebas"] = meta.audio_timebase
        video_meta["audio_numerator"] = meta.audio_timebase.numerator
        video_meta["audio_denominator"] = meta.audio_timebase.denominator
        video_meta["has_audio"] = meta.has_audio
        video_meta["audio_duration"] = meta.audio_duration
        video_meta["audio_sample_rate"] = meta.audio_sample_rate

    if (
        video_meta["has_video"] and
        video_meta["video_denominator"] > 0 and
        video_meta["video_duration"] > 0
    ):
        decode_all_video = False
        start_idx, end_idx = get_start_end_idx_centered(
            video_meta["video_fps"] * video_meta["video_duration"],
            sampling_rate * num_frames / target_fps * video_meta["video_fps"],
            clip_idx,
            num_clips,
        )
        # Convert frame index to pts.
        pts_per_frame = video_meta["video_denominator"] / video_meta["video_fps"]
        video_start_pts = int(start_idx * pts_per_frame)
        video_end_pts = int(end_idx * pts_per_frame)

    # Decode the raw video with the tv decoder.
    v_frames, _ = io._read_video_from_memory(
        video_tensor,
        seek_frame_margin=1.0,
        read_video_stream="visual" in modalities,
        video_width=0,
        video_height=0,
        video_min_dimension=max_spatial_scale,
        video_pts_range=(video_start_pts, video_end_pts),
        video_timebase_numerator=video_meta["video_numerator"],
        video_timebase_denominator=video_meta["video_denominator"],
    )
    return v_frames, video_meta["video_fps"], decode_all_video


def pyav_decode(
    container, sampling_rate, num_frames, clip_idx, num_clips=10, target_fps=30
):
    """
    Convert the video from its original fps to the target_fps. If the video
    support selective decoding (contain decoding information in the video head),
    the perform temporal selective decoding and sample a clip from the video
    with the PyAV decoder. If the video does not support selective decoding,
    decode the entire video.

    Args:
        container (container): pyav container.
        sampling_rate (int): frame sampling rate (interval between two sampled
            frames.
        num_frames (int): number of frames to sample.
        clip_idx (int): if clip_idx is -1, perform random temporal sampling. If
            clip_idx is larger than -1, uniformly split the video to num_clips
            clips, and select the clip_idx-th video clip.
        num_clips (int): overall number of clips to uniformly sample from the
            given video.
        target_fps (int): the input video may has different fps, convert it to
            the target video fps before frame sampling.
    Returns:
        frames (tensor): decoded frames from the video. Return None if the no
            video stream was found.
        fps (float): the number of frames per second of the video.
        decode_all_video (bool): If True, the entire video was decoded.
    """
    # Try to fetch the decoding information from the video head. Some of the
    # videos does not support fetching the decoding information, for that case
    # it will get None duration.
    fps = float(container.streams.video[0].average_rate)
    frames_length = container.streams.video[0].frames
    duration = container.streams.video[0].duration

    if duration is None:
        # If failed to fetch the decoding information, decode the entire video.
        decode_all_video = True
        video_start_pts, video_end_pts = 0, math.inf
    else:
        # Perform selective decoding.
        decode_all_video = False
        start_idx, end_idx = get_start_end_idx_centered(
            frames_length,
            sampling_rate * num_frames / target_fps * fps,
            clip_idx,
            num_clips,
        )
        timebase = duration / frames_length
        video_start_pts = int(start_idx * timebase)
        video_end_pts = int(end_idx * timebase)

    frames = None
    # If video stream was found, fetch video frames from the video.
    if container.streams.video:
        video_frames, max_pts = pyav_decode_stream(
            container,
            video_start_pts,
            video_end_pts,
            container.streams.video[0],
            {"video": 0},
        )
        container.close()

        frames = [frame.to_rgb().to_ndarray() for frame in video_frames]
        frames = torch.as_tensor(np.stack(frames))
    return frames, fps, decode_all_video


def decode(
    container,
    sampling_rate,
    num_frames,
    clip_idx=-1,
    num_clips=10,
    video_meta=None,
    target_fps=30,
    backend="pyav",
    max_spatial_scale=0,
):
    """
    Decode the video and perform temporal sampling.
    Args:
        container (container): pyav container.
        sampling_rate (int): frame sampling rate (interval between two sampled
            frames).
        num_frames (int): number of frames to sample.
        clip_idx (int): if clip_idx is -1, perform random temporal
            sampling. If clip_idx is larger than -1, uniformly split the
            video to num_clips clips, and select the
            clip_idx-th video clip.
        num_clips (int): overall number of clips to uniformly
            sample from the given video.
        video_meta (dict): a dict contains VideoMetaData. Details can be find
            at `pytorch/vision/torchvision/io/_video_opt.py`.
        target_fps (int): the input video may have different fps, convert it to
            the target video fps before frame sampling.
        backend (str): decoding backend includes `pyav` and `torchvision`. The
            default one is `pyav`.
        max_spatial_scale (int): keep the aspect ratio and resize the frame so
            that shorter edge size is max_spatial_scale. Only used in
            `torchvision` backend.
    Returns:
        frames (tensor): decoded frames from the video.
    """
    # Currently support two decoders: 1) PyAV, and 2) TorchVision.
    assert clip_idx >= -1, "Not valied clip_idx {}".format(clip_idx)
    try:
        if backend == "pyav":
            frames, fps, decode_all_video = pyav_decode(
                container, sampling_rate, num_frames, clip_idx, num_clips, target_fps,
            )
        elif backend == "torchvision":
            frames, fps, decode_all_video = torchvision_decode(
                container,
                sampling_rate,
                num_frames,
                clip_idx,
                video_meta,
                num_clips,
                target_fps,
                ("visual",),
                max_spatial_scale,
            )
        else:
            raise NotImplementedError("Unknown decoding backend {}".format(backend))
    except Exception as e:
        logger.debug("Failed to decode by {} with exception: {}".format(backend, e))
        return None

    # Return None if the frames was not decoded successfully.
    if frames is None or frames.size(0) == 0:
        return None

    start_idx, end_idx = get_start_end_idx(
        frames.shape[0],
        num_frames * sampling_rate * fps / target_fps,
        clip_idx if decode_all_video else 0,
        num_clips if decode_all_video else 1,
    )
    # Perform temporal sampling from the decoded video.
    frames = temporal_sampling(frames, start_idx, end_idx, num_frames)
    return frames



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
import json
import torch
import yaml
import os
from pathlib import Path
from opendr.engine.learners import Learner
from opendr.engine.helper.io import bump_version
from torch import onnx
import onnxruntime as ort
from opendr.engine.data import Video
from opendr.engine.datasets import Dataset
from opendr.engine.target import Category

from opendr.perception.activity_recognition.x3d.algorithm.x3d import X3D
import pytorch_lightning as pl

from urllib.request import urlretrieve
from logging import getLogger
from typing import Any, Iterable, Union, Dict, List

logger = getLogger(__name__)

_MODEL_NAMES = {"xs", "s", "m", "l"}


class X3DLearner(Learner):
    def __init__(
        self,
        lr=1e-3,
        iters=10,  # Epochs
        batch_size=64,
        optimizer="adam",
        lr_schedule="",
        backbone="s",
        network_head="classification",
        checkpoint_after_iter=0,
        checkpoint_load_iter=0,
        temp_path="",
        device="cuda",
        loss="cross_entropy",
        weight_decay=1e-5,
        momentum=0.9,
        drop_last=True,
        pin_memory=False,
        num_workers=0,
        seed=123,
        num_classes=400,
        *args,
        **kwargs,
    ):
        """Initialise the X3DLearner

        Args:
            lr (float, optional): Learning rate during optimization. Defaults to 1e-3.
            iters (int, optional): Number of epochs to train for. Defaults to 10.
            batch_size (int, optional): Dataloader batch size. Defaults to 64.
            optimizer (str, optional): Name of optimizer to use ("sgd" or "adam"). Defaults to "adam".
            lr_schedule (str, optional): Unused parameter. Defaults to "".
            network_head (str, optional): Head of network (only "classification" is currently available).
                Defaults to "classification".
            checkpoint_after_iter (int, optional): Unused parameter. Defaults to 0.
            checkpoint_load_iter (int, optional): Unused parameter. Defaults to 0.
            temp_path (str, optional): Path in which to store temporary files. Defaults to "".
            device (str, optional): Name of computational device ("cpu" or "cuda"). Defaults to "cuda".
            loss (str, optional): Loss function used during optimization. Defaults to "cross_entropy".
            weight_decay ([type], optional): Weight decay used for optimization. Defaults to 1e-5.
            momentum (float, optional): Momentum used for optimization. Defaults to 0.9.
            drop_last (bool, optional): Drop last data point if a batch cannot be filled. Defaults to True.
            pin_memory (bool, optional): Pin memory in dataloader. Defaults to False.
            num_workers (int, optional): Number of workers in dataloader. Defaults to 0.
            seed (int, optional): Random seed. Defaults to 123.
            num_classes (int, optional): Number of classes to predict among. Defaults to 400.
        """
        assert backbone in _MODEL_NAMES, f"Invalid model selected. Choose one of {_MODEL_NAMES}."
        assert network_head in {"classification"}, "Currently, only 'classification' head is supported."

        assert optimizer in {"sgd", "adam"}, "Supported optimizers are Adam and SGD."

        # Pass the shared parameters on super's constructor so they can get initialized as class attributes
        super(X3DLearner, self).__init__(
            lr=lr,
            iters=iters,
            batch_size=batch_size,
            optimizer=optimizer,
            lr_schedule=lr_schedule,
            backbone=backbone,
            network_head=network_head,
            temp_path=temp_path,
            checkpoint_after_iter=checkpoint_after_iter,
            checkpoint_load_iter=checkpoint_load_iter,
            device=device,
            threshold=0.0,
        )

        self.weight_decay = weight_decay
        self.momentum = momentum
        self.drop_last = drop_last
        self.pin_memory = pin_memory
        self.num_workers = num_workers
        self.seed = seed
        self.num_classes = num_classes
        self.loss = loss
        self._ort_session = None
        torch.manual_seed(self.seed)

        self._load_model_hparams(self.backbone)
        self.init_model()

    def _load_model_hparams(self, model_name: str = None) -> Dict[str, Any]:
        """Load hyperparameters for an X3D model

        Args:
            model_name (str, optional): Name of the model (one of {"xs", "s", "m", "l"}).
                If none, `self.backbone` is used. Defaults to None.

        Returns:
            Dict[str, Any]: Dictionary with model hyperparameters
        """
        model_name = model_name or self.backbone
        assert model_name in _MODEL_NAMES, f"Invalid model selected. Choose one of {_MODEL_NAMES}."
        path = Path(__file__).parent / "hparams" / f"{model_name}.yaml"
        with open(path, "r") as f:
            self.model_hparams = yaml.load(f, Loader=yaml.FullLoader)
        return self.model_hparams

    def _load_model_weights(self, weights_path: Union[str, Path]):
        """Load pretrained model weights

        Args:
            weights_path (Union[str, Path]): Path to model weights file.
                Type of file must be one of {".pyth", ".pth", ".onnx"}
        """
        weights_path = Path(weights_path)

        assert weights_path.is_file() and weights_path.suffix in {".pyth", ".pth", ".onnx"}, (
            f"weights_path ({str(weights_path)}) should be a .pth or .onnx file."
            "Pretrained weights can be downloaded using `self.download(...)`"
        )
        if weights_path.suffix == ".onnx":
            return self._load_onnx(weights_path)

        logger.debug(f"Loading model weights from {str(weights_path)}")

        # Check for configuration mismatches, loading only matching weights
        loaded_state_dict = torch.load(weights_path, map_location=torch.device(self.device))
        if "model_state" in loaded_state_dict:  # As found in the official pretrained X3D models
            loaded_state_dict = loaded_state_dict["model_state"]
        self.model.load_state_dict(loaded_state_dict, strict=False)
        self.model.to(self.device)

        return self

    def init_model(self) -> X3D:
        """Initialise model with random parameters

        Returns:
            X3D: model
        """
        assert hasattr(self, "model_hparams"), "`self.model_hparams` not found. Did you forget to call `_load_hparams`?"
        self.model = X3D(
            dim_in=3,
            image_size=self.model_hparams["image_size"],
            frames_per_clip=self.model_hparams["frames_per_clip"],
            num_classes=self.num_classes,
            conv1_dim=self.model_hparams["conv1_dim"],
            conv5_dim=self.model_hparams["conv5_dim"],
            num_groups=self.model_hparams["num_groups"],
            width_per_group=self.model_hparams["width_per_group"],
            width_factor=self.model_hparams["width_factor"],
            depth_factor=self.model_hparams["depth_factor"],
            bottleneck_factor=self.model_hparams["bottleneck_factor"],
            use_channelwise_3x3x3=self.model_hparams["use_channelwise_3x3x3"],
            dropout_rate=self.model_hparams["dropout_rate"],
            head_activation=self.model_hparams["head_activation"],
            head_batchnorm=self.model_hparams["head_batchnorm"],
            fc_std_init=self.model_hparams["fc_std_init"],
            final_batchnorm_zero_init=self.model_hparams["final_batchnorm_zero_init"],
        ).to(device=self.device)
        return self.model

    def save(self, path: Union[str, Path]):
        """Save model weights and metadata to path.

        Args:
            path (Union[str, Path]): Directory in which to save model weights and meta data.

        Returns:
            self
        """
        assert hasattr(self, "model"), "Cannot save model because no model was found. Did you forget to call `__init__`?"

        root_path = Path(path)
        root_path.mkdir(parents=True, exist_ok=True)
        name = f"x3d_{self.backbone}"
        ext = ".onnx" if self._ort_session else ".pth"
        weights_path = bump_version(root_path / f"model_{name}{ext}")
        meta_path = bump_version(root_path / f"{name}.json")

        logger.info(f"Saving model weights to {str(weights_path)}")
        if self._ort_session:
            self._save_onnx(weights_path)
        else:
            torch.save(self.model.state_dict(), weights_path)

        logger.info(f"Saving meta-data to {str(meta_path)}")
        meta_data = {
            "model_paths": weights_path.name,
            "framework": "pytorch",
            "format": "pth",
            "has_data": False,
            "inference_params": {
                "backbone": self.backbone,
                "network_head": self.network_head,
                "threshold": self.threshold,
            },
            "optimized": bool(self._ort_session),
            "optimizer_info": {
                "lr": self.lr,
                "iters": self.iters,
                "batch_size": self.batch_size,
                "optimizer": self.optimizer,
                "checkpoint_after_iter": self.checkpoint_after_iter,
                "checkpoint_load_iter": self.checkpoint_load_iter,
                "loss": self.loss,
                "weight_decay": self.weight_decay,
                "momentum": self.momentum,
                "drop_last": self.drop_last,
                "pin_memory": self.pin_memory,
                "num_workers": self.num_workers,
                "seed": self.seed,
            },
        }
        if self._ort_session:
            meta_data["format"] = "onnx"

        with open(str(meta_path), "w", encoding="utf-8") as f:
            json.dump(meta_data, f, sort_keys=True, indent=4)

        return self

    def load(self, path: Union[str, Path]):
        """Load model.

        Args:
            path (Union[str, Path]): Path to metadata file in json format or path to model weights

        Returns:
            self
        """
        path = Path(path)

        # Allow direct loading of weights, omitting the metadatafile
        if path.suffix in {".pyth", ".pth", ".onnx"}:
            self._load_model_weights(path)
            return self
        if path.is_dir():
            path = path / f"x3d_{self.backbone}.json"
        assert path.is_file() and path.suffix == ".json", "The provided metadata path should be a .json file"

        logger.debug(f"Loading X3DLearner metadata from {str(path)}")
        with open(path, "r") as f:
            meta_data = json.load(f)

        inference_params = meta_data["inference_params"]
        optimizer_info = meta_data["optimizer_info"]

        self.__init__(
            lr=optimizer_info["lr"],
            iters=optimizer_info["iters"],
            batch_size=optimizer_info["batch_size"],
            optimizer=optimizer_info["optimizer"],
            device=getattr(self, "device", "cpu"),
            threshold=inference_params["threshold"],
            backbone=inference_params["backbone"],
            network_head=inference_params["network_head"],
            loss=optimizer_info["loss"],
            checkpoint_after_iter=optimizer_info["checkpoint_after_iter"],
            checkpoint_load_iter=optimizer_info["checkpoint_load_iter"],
            weight_decay=optimizer_info["weight_decay"],
            momentum=optimizer_info["momentum"],
            drop_last=optimizer_info["drop_last"],
            pin_memory=optimizer_info["pin_memory"],
            num_workers=optimizer_info["num_workers"],
            seed=optimizer_info["seed"],
        )

        weights_path = path.parent / meta_data["model_paths"]
        self._load_model_weights(weights_path)

        return self

    @staticmethod
    def download(path: Union[str, Path], model_names: Iterable[str] = _MODEL_NAMES):
        """Download pretrained X3D models

        Args:
            path (Union[str, Path], optional): Directory in which to store model weights. Defaults to None.
            model_names (Iterable[str], optional): iterable with model names to download.
                The iterable may contain {"xs", "s", "m", "l"}.
                Defaults to _MODEL_NAMES.
        """
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        for m in model_names:
            assert m in _MODEL_NAMES
            filename = path / f"x3d_{m}.pyth"
            if filename.exists():
                logger.info(f"Skipping download of X3D-{m} (already exists at {str(filename)})")
            else:
                logger.info(f"Downloading pretrained X3D-{m} weight to {str(filename)}")
                urlretrieve(
                    url=f"https://dl.fbaipublicfiles.com/pyslowfast/x3d_models/x3d_{m}.pyth",
                    filename=str(filename),
                )
                assert filename.is_file(), f"Something wen't wrong when downloading {str(filename)}"

    def reset(self):
        pass

    def fit(
        self,
        dataset: Dataset,
        val_dataset: Dataset = None,
        epochs: int = None,
        steps: int = None,
        *args,
        **kwargs,
    ):
        """Fit the model to a dataset

        Args:
            dataset (Dataset): Training dataset
            val_dataset (Dataset, optional): Validation dataset.
                If none is given, validation steps are skipped. Defaults to None.
            epochs (int, optional): Number of epochs. If none is supplied, self.iters will be used. Defaults to None.
            steps (int, optional): Number of training steps to conduct. If none, this is determined by epochs. Defaults to None.
        """

        train_dataloader = torch.utils.data.DataLoader(
            dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=True,
            pin_memory=self.pin_memory,
            drop_last=self.drop_last,
        )
        val_dataloader = (
            torch.utils.data.DataLoader(
                val_dataset,
                batch_size=self.batch_size,
                num_workers=self.num_workers,
                shuffle=False,
                pin_memory=self.pin_memory,
                drop_last=self.drop_last,
            )
            if val_dataset
            else None
        )

        optimisation_metric = "val/loss" if val_dataset else "train/loss"

        # Patch model optimizer
        assert self.optimizer in {
            "adam",
            "sgd",
        }, f"Invalid optimizer '{self.optimizer}'. Must be 'adam' or 'sgd'."
        if self.optimizer == "adam":
            Optimizer = partial(
                torch.optim.Adam,
                lr=self.lr,
                betas=(self.momentum, 0.999),
                weight_decay=self.weight_decay,
            )
        else:  # self.optimizer == "sgd":
            Optimizer = partial(
                torch.optim.Adam,
                lr=self.lr,
                momentum=self.momentum,
                weight_decay=self.weight_decay,
            )

        def configure_optimizers():
            # nonlocal Optimizer, optimisation_metric
            optimizer = Optimizer(self.model.parameters())
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, patience=10)
            return {
                "optimizer": optimizer,
                "lr_scheduler": scheduler,
                "monitor": optimisation_metric,
            }

        model = getattr(self, "_plmodel", self.model)
        model.configure_optimizers = configure_optimizers

        self.trainer = pl.Trainer(
            max_epochs=epochs or self.iters,
            gpus=1 if "cuda" in self.device else 0,
            callbacks=[
                pl.callbacks.ModelCheckpoint(
                    save_top_k=1,
                    verbose=True,
                    monitor=optimisation_metric,
                    mode="min",
                    prefix="",
                )
            ],
            logger=_experiment_logger(),
        )
        self.trainer.limit_train_batches = steps or self.trainer.limit_train_batches
        self.trainer.limit_val_batches = steps or self.trainer.limit_val_batches

        self.trainer.fit(model, train_dataloader, val_dataloader)
        self.model.to(self.device)

    def eval(self, dataset: Dataset, steps: int = None) -> Dict[str, Any]:
        """Evaluate the model on the dataset

        Args:
            dataset (Dataset): Dataset on which to evaluate model
            steps (int, optional): Number of validation batches to evaluate.
                If None, all batches are evaluated. Defaults to None.

        Returns:
            Dict[str, Any]: Evaluation statistics
        """
        test_dataloader = torch.utils.data.DataLoader(
            dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=False,
            pin_memory=self.pin_memory,
            drop_last=False,
        )

        if not hasattr(self, "trainer"):
            self.trainer = pl.Trainer(
                gpus=1 if "cuda" in self.device else 0,
                logger=_experiment_logger(),
            )
        self.trainer.limit_test_batches = steps or self.trainer.limit_test_batches
        model = getattr(self, "_plmodel", self.model)
        results = self.trainer.test(model, test_dataloader)
        results = {
            "accuracy": results[-1]["test/acc"],
            "loss": results[-1]["test/loss"],
        }
        return results

    def infer(self, batch: Union[Video, List[Video], torch.Tensor]) -> List[Category]:
        """Run inference on a batch of data

        Args:
            batch (torch.Tensor): Video or batch of video-clips.
                The video should have shape (3, T, H, W). If a batch is supplied, its shape should be (B, 3, T, H, W).

        Returns:
            torch.Tensor: Network output
        """
        # Cast to torch tensor
        if type(batch) is Video:
            batch = [batch]
        if type(batch) is list:
            batch = torch.stack([torch.tensor(v.data) for v in batch])

        batch = batch.to(device=self.device, dtype=torch.float)
        if self._ort_session is not None:
            results = torch.tensor(self._ort_session.run(None, {"video": batch.cpu().numpy()})[0])
        else:
            self.model.eval()
            results = self.model.forward(batch)
        results = [Category(prediction=int(r.argmax(dim=0)), confidence=r) for r in results]
        return results

    def optimize(self, do_constant_folding=False):
        """Optimize model execution.
        This is accomplished by saving to the ONNX format and loading the optimized model.

        Args:
            do_constant_folding (bool, optional): Whether to optimize constants. Defaults to False.
        """

        if getattr(self.model, "_ort_session", None):
            logger.info("Model is already optimized. Skipping redundant optimization")
            return

        path = Path(self.temp_path or os.getcwd()) / "weights" / f"x3d_{self.backbone}.onnx"
        if not path.exists():
            self._save_onnx(path, do_constant_folding)
        self._load_onnx(path)

    @property
    def _example_input(self):
        C = 3  # RGB
        T = self.model_hparams["frames_per_clip"]
        S = self.model_hparams["image_size"]
        return torch.randn(1, C, T, S, S).to(device=self.device)

    @_example_input.setter
    def _example_input(self):
        raise ValueError("_example_input is set thorugh 'frames_per_clip' 'image_size' in `self.model_hparams`")

    def _save_onnx(self, path: Union[str, Path], do_constant_folding=False, verbose=False):
        """Save model in the ONNX format

        Args:
            path (Union[str, Path]): Directory in which to save ONNX model
            do_constant_folding (bool, optional): Whether to optimize constants. Defaults to False.
        """
        path.parent.mkdir(exist_ok=True, parents=True)

        self.model.eval()
        self.model.to(device=self.device)

        logger.info(f"Saving model to ONNX format at {str(path)}")
        onnx.export(
            self.model,
            self._example_input,
            path,
            input_names=["video"],
            output_names=["classes"],
            dynamic_axes={"video": {0: "batch_size"}, "classes": {0: "batch_size"}},
            do_constant_folding=do_constant_folding,
            verbose=verbose,
            opset_version=11,
        )

    def _load_onnx(self, path: Union[str, Path]):
        """Loads ONNX model into an onnxruntime inference session.

        Args:
            path (Union[str, Path]): Path to ONNX model
        """
        logger.info(f"Loading ONNX runtime inference session from {str(path)}")
        self._ort_session = ort.InferenceSession(str(path))


def _experiment_logger():
    return pl.loggers.TensorBoardLogger(save_dir=Path(os.getcwd()) / "logs", name="x3d")



""" Adapted from: https://github.com/facebookresearch/SlowFast
"""

import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from .head_helper import X3DHead
from .resnet_helper import ResStage
from .stem_helper import VideoModelStem
import pytorch_lightning as pl


class X3D(pl.LightningModule):
    """
    X3D model,
    adapted from https://github.com/facebookresearch/SlowFast

    Christoph Feichtenhofer.
    "X3D: Expanding Architectures for Efficient Video Recognition."
    https://arxiv.org/abs/2004.04730
    """

    def __init__(
        self,
        dim_in: int,
        image_size: int,
        frames_per_clip: int,
        num_classes: int,
        conv1_dim: int,
        conv5_dim: int,
        num_groups: int,
        width_per_group: int,
        width_factor: float,
        depth_factor: float,
        bottleneck_factor: float,
        use_channelwise_3x3x3: bool,
        dropout_rate: float,
        head_activation: str,
        head_batchnorm: bool,
        fc_std_init: float,
        final_batchnorm_zero_init: bool,
        loss_name="cross_entropy",
    ):
        super().__init__()
        self.norm_module = torch.nn.BatchNorm3d
        self.loss_name = loss_name

        exp_stage = 2.0
        self.dim_conv1 = conv1_dim

        self.dim_res2 = (
            _round_width(self.dim_conv1, exp_stage, divisor=8)
            if False  # hparams.X3D.SCALE_RES2
            else self.dim_conv1
        )
        self.dim_res3 = _round_width(self.dim_res2, exp_stage, divisor=8)
        self.dim_res4 = _round_width(self.dim_res3, exp_stage, divisor=8)
        self.dim_res5 = _round_width(self.dim_res4, exp_stage, divisor=8)

        self.block_basis = [
            # blocks, c, stride
            [1, self.dim_res2, 2],
            [2, self.dim_res3, 2],
            [5, self.dim_res4, 2],
            [3, self.dim_res5, 2],
        ]

        num_groups = num_groups
        width_per_group = width_per_group
        dim_inner = num_groups * width_per_group

        w_mul = width_factor
        d_mul = depth_factor
        dim_res1 = _round_width(self.dim_conv1, w_mul)

        # Basis of temporal kernel sizes for each of the stage.
        temp_kernel = [
            [[5]],  # conv1 temporal kernels.
            [[3]],  # res2 temporal kernels.
            [[3]],  # res3 temporal kernels.
            [[3]],  # res4 temporal kernels.
            [[3]],  # res5 temporal kernels.
        ]

        self.s1 = VideoModelStem(
            dim_in=[dim_in],
            dim_out=[dim_res1],
            kernel=[temp_kernel[0][0] + [3, 3]],
            stride=[[1, 2, 2]],
            padding=[[temp_kernel[0][0][0] // 2, 1, 1]],
            norm_module=self.norm_module,
            stem_func_name="x3d_stem",
        )

        # blob_in = s1
        dim_in = dim_res1
        dim_out = dim_in
        for stage, block in enumerate(self.block_basis):
            dim_out = _round_width(block[1], w_mul)
            dim_inner = int(bottleneck_factor * dim_out)

            n_rep = _round_repeats(block[0], d_mul)
            prefix = "s{}".format(stage + 2)  # start w res2 to follow convention

            s = ResStage(
                dim_in=[dim_in],
                dim_out=[dim_out],
                dim_inner=[dim_inner],
                temp_kernel_sizes=temp_kernel[1],
                stride=[block[2]],
                num_blocks=[n_rep],
                num_groups=[dim_inner] if use_channelwise_3x3x3 else [num_groups],
                num_block_temp_kernel=[n_rep],
                nonlocal_inds=[[]],
                nonlocal_group=[1],
                nonlocal_pool=[[1, 2, 2], [1, 2, 2]],
                instantiation="dot_product",
                trans_func_name="x3d_transform",
                stride_1x1=False,
                norm_module=self.norm_module,
                dilation=[1],
                drop_connect_rate=0.0,
            )
            dim_in = dim_out
            self.add_module(prefix, s)

        spat_sz = int(math.ceil(image_size / 32.0))
        self.head = X3DHead(
            dim_in=dim_out,
            dim_inner=dim_inner,
            dim_out=conv5_dim,
            num_classes=num_classes,
            pool_size=(frames_per_clip, spat_sz, spat_sz),
            dropout_rate=dropout_rate,
            act_func=head_activation,
            bn_lin5_on=bool(head_batchnorm),
        )
        init_weights(self, fc_std_init, bool(final_batchnorm_zero_init))

    def forward(self, x: Tensor):
        # The original slowfast code was set up to use multiple paths, wrap the input
        x = [x]  # type:ignore
        for module in self.children():
            x = module(x)
        return x

    def training_step(self, batch, batch_idx):
        x, y = batch
        x = self.forward(x)
        loss = getattr(F, self.loss_name, F.cross_entropy)(x, y)
        self.log('train/loss', loss)
        self.log('train/acc', _accuracy(x, y))
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        x = self.forward(x)
        loss = getattr(F, self.loss_name, F.cross_entropy)(x, y)
        self.log('val/loss', loss)
        self.log('val/acc', _accuracy(x, y))
        return loss

    def test_step(self, batch, batch_idx):
        x, y = batch
        x = self.forward(x)
        loss = getattr(F, self.loss_name, F.cross_entropy)(x, y)
        self.log('test/loss', loss)
        self.log('test/acc', _accuracy(x, y))
        return loss


def _accuracy(x: Tensor, y: Tensor):
    return torch.sum(x.argmax(dim=1) == y) / len(y)


def _round_width(width, multiplier, min_depth=8, divisor=8):
    """Round width of filters based on width multiplier."""
    if not multiplier:
        return width

    width *= multiplier
    min_depth = min_depth or divisor
    new_filters = max(min_depth, int(width + divisor / 2) // divisor * divisor)
    if new_filters < 0.9 * width:
        new_filters += divisor
    return int(new_filters)


def _round_repeats(repeats, multiplier):
    """Round number of layers based on depth multiplier."""
    multiplier = multiplier
    if not multiplier:
        return repeats
    return int(math.ceil(multiplier * repeats))


def c2_msra_fill(module: nn.Module) -> None:
    """
    Initialize `module.weight` using the "MSRAFill" implemented in Caffe2.
    Also initializes `module.bias` to 0.

    Args:
        module (torch.nn.Module): module to initialize.
    """
    # pyre-ignore
    nn.init.kaiming_normal_(module.weight, mode="fan_out", nonlinearity="relu")
    if module.bias is not None:  # pyre-ignore
        nn.init.constant_(module.bias, 0)


def init_weights(model, fc_init_std=0.01, zero_init_final_bn=True):
    """
    Performs ResNet style weight initialization.
    Args:
        fc_init_std (float): the expected standard deviation for fc layer.
        zero_init_final_bn (bool): if True, zero initialize the final bn for
            every bottleneck.
    """
    for m in model.modules():
        if isinstance(m, nn.Conv3d):
            """
            Follow the initialization method proposed in:
            {He, Kaiming, et al.
            "Delving deep into rectifiers: Surpassing human-level
            performance on imagenet classification."
            arXiv preprint arXiv:1502.01852 (2015)}
            """
            c2_msra_fill(m)
        elif isinstance(m, nn.BatchNorm3d):
            if (
                hasattr(m, "transform_final_bn") and
                m.transform_final_bn and
                zero_init_final_bn
            ):
                batchnorm_weight = 0.0
            else:
                batchnorm_weight = 1.0
            if m.weight is not None:
                m.weight.data.fill_(batchnorm_weight)
            if m.bias is not None:
                m.bias.data.zero_()
        if isinstance(m, nn.Linear):
            m.weight.data.normal_(mean=0.0, std=fc_init_std)
            if m.bias is not None:
                m.bias.data.zero_()



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn.functional as F
import continual as co
import pickle
import os
from opendr.engine import data
from opendr.engine.target import Category
from opendr.perception.activity_recognition.cox3d.algorithm.x3d import CoX3D
from opendr.perception.activity_recognition.utils.lightning import _LightningModuleWithCrossEntropy
from opendr.perception.activity_recognition.x3d.x3d_learner import X3DLearner
from pathlib import Path
from logging import getLogger
from typing import Union, List
import onnxruntime as ort


logger = getLogger(__name__)


class CoX3DLearner(X3DLearner):
    def __init__(
        self,
        lr=1e-3,
        iters=10,  # Epochs
        batch_size=64,
        optimizer="adam",
        lr_schedule="",
        backbone="s",
        network_head="classification",
        checkpoint_after_iter=0,
        checkpoint_load_iter=0,
        temp_path="",
        device="cuda",
        loss="cross_entropy",
        weight_decay=1e-5,
        momentum=0.9,
        drop_last=True,
        pin_memory=False,
        num_workers=0,
        seed=123,
        num_classes=400,
        temporal_window_size: int = None,
        *args,
        **kwargs,
    ):
        """Initialise the CoX3DLearner
        This learner wraps the Continual version of X3D, which makes predictions frame-by-frame rather than by clip.

        Args:
            lr (float, optional): Learning rate during optimization. Defaults to 1e-3.
            iters (int, optional): Number of epochs to train for. Defaults to 10.
            optimizer (str, optional): Name of optimizer to use ("sgd" or "adam"). Defaults to "adam".
            lr_schedule (str, optional): Unused parameter. Defaults to "".
            network_head (str, optional): Head of network (only "classification" is currently available).
                Defaults to "classification".
            checkpoint_after_iter (int, optional): Unused parameter. Defaults to 0.
            checkpoint_load_iter (int, optional): Unused parameter. Defaults to 0.
            temp_path (str, optional): Path in which to store temporary files. Defaults to "".
            device (str, optional): Name of computational device ("cpu" or "cuda"). Defaults to "cuda".
            weight_decay (float, optional): Weight decay used for optimization. Defaults to 1e-5.
            momentum (float, optional): Momentum used for optimization. Defaults to 0.9.
            drop_last (bool, optional): Drop last data point if a batch cannot be filled. Defaults to True.
            pin_memory (bool, optional): Pin memory in dataloader. Defaults to False.
            loss (str): Name of loss in torch.nn.functional to use. Defaults to "cross_entropy".
            num_workers (int, optional): Number of workers in dataloader. Defaults to 0.
            seed (int, optional): Random seed. Defaults to 123.
            num_classes (int, optional): Number of classes to predict among. Defaults to 400.
            temporal_window_size (int, optional): Size of the final global average pooling.
                If None, size will be automatically chosen according to the backbone. Defaults to None.
        """
        super().__init__(
            lr,
            iters,
            batch_size,
            optimizer,
            lr_schedule,
            backbone,
            network_head,
            checkpoint_after_iter,
            checkpoint_load_iter,
            temp_path,
            device,
            loss,
            weight_decay,
            momentum,
            drop_last,
            pin_memory,
            num_workers,
            seed,
            num_classes,
            *args,
            **kwargs,
        )
        self.temporal_window_size = temporal_window_size
        self._ort_state = None

    def init_model(self) -> CoX3D:
        """Initialise model with random parameters

        Returns:
            CoX3D: model
        """
        assert hasattr(self, "model_hparams"), "`self.model_hparams` not found. Did you forget to call `_load_hparams`?"
        self.model = CoX3D(
            dim_in=3,
            image_size=self.model_hparams["image_size"],
            temporal_window_size=getattr(self, "temporal_window_size", None) or self.model_hparams["frames_per_clip"],
            num_classes=self.num_classes,
            x3d_conv1_dim=self.model_hparams["conv1_dim"],
            x3d_conv5_dim=self.model_hparams["conv5_dim"],
            x3d_num_groups=self.model_hparams["num_groups"],
            x3d_width_per_group=self.model_hparams["width_per_group"],
            x3d_width_factor=self.model_hparams["width_factor"],
            x3d_depth_factor=self.model_hparams["depth_factor"],
            x3d_bottleneck_factor=self.model_hparams["bottleneck_factor"],
            x3d_use_channelwise_3x3x3=self.model_hparams["use_channelwise_3x3x3"],
            x3d_dropout_rate=self.model_hparams["dropout_rate"],
            x3d_head_activation=self.model_hparams["head_activation"],
            x3d_head_batchnorm=self.model_hparams["head_batchnorm"],
            x3d_fc_std_init=self.model_hparams["fc_std_init"],
            x3d_final_batchnorm_zero_init=self.model_hparams["final_batchnorm_zero_init"],
        ).to(device=self.device)
        self._plmodel = _LightningModuleWithCrossEntropy(self.model)
        return self.model

    def _map_state_dict(self, sd):
        if len(sd["head.lin_5.weight"]) > 3:
            sd["head.lin_5.weight"] = sd["head.lin_5.weight"].squeeze(-1).squeeze(-1)
        return sd

    @property
    def _example_input(self):
        C = 3  # RGB
        S = self.model_hparams["image_size"]
        return torch.randn(1, C, S, S).to(device=self.device)

    def infer(self, batch: Union[data.Image, List[data.Image], torch.Tensor]) -> List[Category]:
        """Run inference on a batch of data

        Args:
            batch (torch.Tensor): Image or batch of images.
                The image should have shape (3, H, W). If a batch is supplied, its shape should be (B, 3, H, W).

        Returns:
            List[target.Category]: List of output categories
        """
        # Cast to torch tensor
        if type(batch) is data.Image:
            batch = [batch]
        if type(batch) is list:
            batch = torch.stack([torch.tensor(v.data) for v in batch])

        batch = batch.to(device=self.device, dtype=torch.float)

        if self._ort_session is not None and self._ort_state is not None:
            inputs = {
                "input": batch.cpu().detach().numpy(),
                **self._ort_state,
            }
            results, *next_state = self._ort_session.run(None, inputs)
            results = torch.tensor(results)
            self._ort_state = {k: v for k, v in zip(self._ort_state.keys(), next_state)}
        else:
            self.model.eval()
            results = self.model.forward_step(batch)
        if results is not None:
            results = [Category(prediction=int(r.argmax(dim=0)), confidence=F.softmax(r, dim=-1)) for r in results]
        return results

    def optimize(self, do_constant_folding=False):
        """Optimize model execution.
        This is accomplished by saving to the ONNX format and loading the optimized model.

        Args:
            do_constant_folding (bool, optional): Whether to optimize constants. Defaults to False.
        """

        if getattr(self.model, "_ort_session", None):
            logger.info("Model is already optimized. Skipping redundant optimization")
            return

        path = Path(self.temp_path or os.getcwd()) / "weights" / f"cox3d_{self.backbone}.onnx"
        if not path.exists():
            self._save_onnx(path, do_constant_folding)
        self._load_onnx(path)

    def _save_onnx(self, path: Union[str, Path], do_constant_folding=False, verbose=False):
        """Save model in the ONNX format

        Args:
            path (Union[str, Path]): Directory in which to save ONNX model
            do_constant_folding (bool, optional): Whether to optimize constants. Defaults to False.
        """
        path.parent.mkdir(exist_ok=True, parents=True)

        model = self.model.to(device="cpu")
        model.eval()

        # Prepare state
        state0 = None
        sample = self._example_input.repeat(self.batch_size, 1, 1, 1)
        with torch.no_grad():
            for _ in range(model.receptive_field):
                _, state0 = model._forward_step(sample, state0)
            _, state0 = model._forward_step(sample, state0)
            state0 = co.utils.flatten(state0)

            # Export to ONNX
            logger.info(f"Saving model to ONNX format at {str(path)}")
            co.onnx.export(
                model,
                (sample, *state0),
                path,
                input_names=["input"],
                output_names=["output"],
                do_constant_folding=do_constant_folding,
                verbose=verbose,
                opset_version=11,
            )

        # Save default state and name mappings for later use
        state_path = path.parent / f"cox3d_{self.backbone}_state.pickle"
        logger.info(f"Saving ONNX model states at {str(state_path)}")
        omodel = co.onnx.OnnxWrapper(self.model)
        state = {k: v.detach().numpy() for k, v in zip(omodel.state_input_names, state0)}
        with open(state_path, "wb") as f:
            pickle.dump(state, f)

    def _load_onnx(self, path: Union[str, Path]):
        """Loads ONNX model into an onnxruntime inference session.

        Args:
            path (Union[str, Path]): Path to ONNX model
        """
        onnx_path = path
        state_path = path.parent / f"cox3d_{self.backbone}_state.pickle"

        logger.info(f"Loading ONNX runtime inference session from {str(onnx_path)}")
        self._ort_session = ort.InferenceSession(str(onnx_path))

        logger.info(f"Loading ONNX state from {str(state_path)}")
        with open(state_path, "rb") as f:
            self._ort_state = pickle.load(f)

    def _load_model_weights(self, weights_path: Union[str, Path]):
        """Load pretrained model weights

        Args:
            weights_path (Union[str, Path]): Path to model weights file.
                Type of file must be one of {".pyth", ".pth", ".onnx"}
        """
        weights_path = Path(weights_path)

        assert weights_path.is_file() and weights_path.suffix in {".pyth", ".pth", ".onnx"}, (
            f"weights_path ({str(weights_path)}) should be a .pth or .onnx file."
            "Pretrained weights can be downloaded using `self.download(...)`"
        )
        if weights_path.suffix == ".onnx":
            return self._load_onnx(weights_path)

        logger.debug(f"Loading model weights from {str(weights_path)}")

        # Check for configuration mismatches, loading only matching weights
        loaded_state_dict = torch.load(weights_path, map_location=torch.device(self.device))
        if "model_state" in loaded_state_dict:  # As found in the official pretrained X3D models
            loaded_state_dict = loaded_state_dict["model_state"]

        loaded_state_dict = self._map_state_dict(loaded_state_dict)

        self.model.load_state_dict(loaded_state_dict, strict=False, flatten=True)
        self.model.to(self.device)

        return self



from opendr.perception.activity_recognition.datasets.dummy_timeseries_dataset import (
    DummyTimeseriesDataset,
)
from opendr.perception.activity_recognition.datasets.kinetics import (
    KineticsDataset,
    CLASSES as KINETICS_CLASSES,
)

__all__ = [
    "KineticsDataset",
    "KINETICS_CLASSES",
    "DummyTimeseriesDataset",
]



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
import json
import torch
import os
import pickle
from pathlib import Path
from opendr.engine.learners import Learner
from opendr.engine.helper.io import bump_version
import onnxruntime as ort
from collections import OrderedDict

from opendr.engine.data import Timeseries, Vector
from opendr.engine.datasets import Dataset
from opendr.engine.target import Category

from logging import getLogger
from typing import Any, Union, Dict

import pytorch_lightning as pl
import continual as co
from continual import onnx

logger = getLogger(__name__)


class CoTransEncLearner(Learner):
    def __init__(
        self,
        lr=1e-2,
        iters=10,  # Epochs
        batch_size=64,
        optimizer="sgd",
        lr_schedule="",
        network_head="classification",
        num_layers=1,  # 1 or 2
        input_dims=1024,
        hidden_dims=1024,
        sequence_len=64,
        num_heads=8,
        dropout=0.1,
        num_classes=22,
        positional_encoding_learned=False,
        checkpoint_after_iter=0,
        checkpoint_load_iter=0,
        temp_path="",
        device="cuda",
        loss="cross_entropy",
        weight_decay=1e-4,
        momentum=0.9,
        drop_last=True,
        pin_memory=False,
        num_workers=0,
        seed=123,
        *args,
        **kwargs,
    ):
        """Initialise the CoTransLearner
        This learner wraps the Continual Transformer Encoder with Recycling Positional Encoding
        which optimises token-by-token predictions for temporal sequences.
        It was proposed for Online Action Detection in
        "L. Hedegaard, A. Bakhtiarnia, and A. Iosifidis: 'Continual Transformers: Redundancy-Free
        Attention for Online Inference', 2022"
        https://arxiv.org/abs/2201.06268

        Args:
            lr (float, optional): Learning rate during optimization. Defaults to 1e-3.
            iters (int, optional): Number of epochs to train for. Defaults to 10.
            optimizer (str, optional): Name of optimizer to use ("sgd" or "adam"). Defaults to "adam".
            lr_schedule (str, optional): Schedule for training the model. Only "ReduceLROnPlateau" is available currently.
            network_head (str, optional): Head of network (only "classification" is currently available).
                Defaults to "classification".
            num_layers (int, optional): Number of Transformer Encoder layers (1 or 2). Defaults to 1.
            input_dims (int, optional): Input token dimension. Defaults to 1024.
            hidden_dims (int, optional): Hidden projection dimension. Defaults to 1024.
            sequence_len (int, optional): Length of token sequence to consider. Defaults to 64.
            num_heads (int, optional): Number of attention heads. Defaults to 8.
            dropout (float, optional): Dropout probability. Defaults to 0.1.
            num_classes (int, optional): Number of classes to predict. Defaults to 22.
            positional_encoding_learned (bool, optional): Whether positional encoding is learned. Defaults to False.
            checkpoint_after_iter (int, optional): Unused parameter. Defaults to 0.
            checkpoint_load_iter (int, optional): Unused parameter. Defaults to 0.
            temp_path (str, optional): Path in which to store temporary files. Defaults to "".
            device (str, optional): Name of computational device ("cpu" or "cuda"). Defaults to "cuda".
            weight_decay ([type], optional): Weight decay used for optimization. Defaults to 1e-5.
            momentum (float, optional): Momentum used for optimization. Defaults to 0.9.
            drop_last (bool, optional): Drop last data point if a batch cannot be filled. Defaults to True.
            pin_memory (bool, optional): Pin memory in dataloader. Defaults to False.
            num_workers (int, optional): Number of workers in dataloader. Defaults to 0.
            seed (int, optional): Random seed. Defaults to 123.
        """
        # Pass the shared parameters on super's constructor so they can get initialized as class attributes
        assert optimizer in {"sgd", "adam"}, "Supported optimizers are Adam and SGD."
        assert network_head in {
            "classification"
        }, "Currently, only 'classification' head is supported."

        super(CoTransEncLearner, self).__init__(
            lr=lr,
            iters=iters,
            batch_size=batch_size,
            optimizer=optimizer,
            lr_schedule=lr_schedule,
            network_head=network_head,
            temp_path=temp_path,
            checkpoint_after_iter=checkpoint_after_iter,
            checkpoint_load_iter=checkpoint_load_iter,
            device=device,
        )

        assert num_layers in {
            1,
            2,
        }, "Only 1 or 2 Transformer Encoder layers are supported."
        self._num_layers = num_layers
        self._positional_encoding_learned = positional_encoding_learned
        self._input_dims = input_dims
        self._hidden_dims = hidden_dims
        self._sequence_len = sequence_len
        self._num_heads = num_heads
        self._dropout = dropout
        self._num_classes = num_classes
        self._weight_decay = weight_decay
        self._momentum = momentum
        self._drop_last = drop_last
        self._pin_memory = pin_memory
        self._num_workers = num_workers
        self._loss = loss
        self._ort_session = None
        self._seed = seed
        torch.manual_seed(self._seed)

        self.init_model()

    def init_model(self) -> torch.nn.Module:
        """Initialise model with random parameters

        Returns:
            torch.nn.Module: Continual Transformer Encoder with
                Recycling Positional Encoding
        """
        pos_enc = co.RecyclingPositionalEncoding(
            embed_dim=self._input_dims,
            num_embeds=self._sequence_len * 2 - 1,
            learned=self._positional_encoding_learned,
        )
        trans_enc = co.TransformerEncoder(
            co.TransformerEncoderLayerFactory(
                d_model=self._input_dims,
                nhead=self._num_heads,
                dim_feedforward=self._hidden_dims,
                dropout=self._dropout,
                sequence_len=self._sequence_len,
            ),
            num_layers=self._num_layers,
        )
        lin = co.Linear(self._input_dims, self._num_classes, channel_dim=-1)

        self.model = co.Sequential(
            OrderedDict(
                [
                    ("pos_enc", pos_enc),
                    ("trans_enc", trans_enc),
                    (
                        "select",
                        co.Lambda(
                            fn=lambda x: x[:, :, -1],
                            forward_step_only_fn=lambda x: x,
                            takes_time=True,
                        ),
                    ),
                    ("lin", lin),
                ]
            )
        )

        class AddIfTraining:
            def __init__(slf, val: int):
                slf.val = val

            def __add__(slf, other: int):
                return other + (slf.val if self.model.training else 0)

            def __radd__(slf, other: int):
                return slf.__add__(other)

        self.model[0].forward_update_index_steps = AddIfTraining(1)
        self.model = self.model.to(device=self.device)

        self._plmodel = _LightningModuleWithCrossEntropy(self.model)
        return self.model

    def save(self, path: Union[str, Path]):
        """Save model weights and metadata to path.

        Args:
            path (Union[str, Path]): Directory in which to save model weights and meta data.

        Returns:
            self
        """
        assert hasattr(
            self, "model"
        ), "Cannot save model because no model was found. Did you forget to call `__init__`?"

        root_path = Path(path)
        root_path.mkdir(parents=True, exist_ok=True)
        name = "cotransenc_weights"
        ext = ".onnx" if self._ort_session else ".pth"
        weights_path = bump_version(root_path / f"model_{name}{ext}")
        meta_path = bump_version(root_path / f"{name}.json")

        logger.info(f"Saving model weights to {str(weights_path)}")
        if self._ort_session:
            self._save_onnx(weights_path)
        else:
            torch.save(self.model.state_dict(), weights_path)

        logger.info(f"Saving meta-data to {str(meta_path)}")
        meta_data = {
            "model_paths": weights_path.name,
            "framework": "pytorch",
            "format": "pth",
            "has_data": False,
            "inference_params": {
                "network_head": self._network_head,
                "num_layers": self._num_layers,
                "input_dims": self._input_dims,
                "hidden_dims": self._hidden_dims,
                "sequence_len": self._sequence_len,
                "num_heads": self._num_heads,
                "dropout": self._dropout,
                "num_classes": self._num_classes,
                "positional_encoding_learned": self._positional_encoding_learned,
            },
            "optimized": bool(self._ort_session),
            "optimizer_info": {
                "lr": self.lr,
                "iters": self.iters,
                "batch_size": self.batch_size,
                "optimizer": self.optimizer,
                "checkpoint_after_iter": self.checkpoint_after_iter,
                "checkpoint_load_iter": self.checkpoint_load_iter,
                "loss": self._loss,
                "weight_decay": self._weight_decay,
                "momentum": self._momentum,
                "drop_last": self._drop_last,
                "pin_memory": self._pin_memory,
                "num_workers": self._num_workers,
                "seed": self._seed,
                "dropout": self._dropout,
            },
        }
        with open(str(meta_path), "w", encoding="utf-8") as f:
            json.dump(meta_data, f, sort_keys=True, indent=4)

        return self

    def load(self, path: Union[str, Path]):
        """Load model.

        Args:
            path (Union[str, Path]): Path to metadata file in json format or path to model weights

        Returns:
            self
        """
        path = Path(path)

        # Allow direct loading of weights, omitting the metadatafile
        if path.suffix in {".pyth", ".pth", ".onnx"}:
            self._load_model_weights(path)
            return self
        if path.is_dir():
            path = path / "cotransenc_weights.json"
        assert (
            path.is_file() and path.suffix == ".json"
        ), "The provided metadata path should be a .json file"

        logger.debug(f"Loading CoTransEnc metadata from {str(path)}")
        with open(path, "r") as f:
            meta_data = json.load(f)

        inference_params = meta_data["inference_params"]
        optimizer_info = meta_data["optimizer_info"]

        self.__init__(
            lr=optimizer_info["lr"],
            iters=optimizer_info["iters"],
            batch_size=optimizer_info["batch_size"],
            optimizer=optimizer_info["optimizer"],
            device=getattr(self, "device", "cpu"),
            network_head=inference_params["network_head"],
            num_layers=inference_params["num_layers"],
            input_dims=inference_params["input_dims"],
            hidden_dims=inference_params["hidden_dims"],
            sequence_len=inference_params["sequence_len"],
            num_heads=inference_params["num_heads"],
            num_classes=inference_params["num_classes"],
            positional_encoding_learned=inference_params["positional_encoding_learned"],
            dropout=optimizer_info["dropout"],
            loss=optimizer_info["loss"],
            checkpoint_after_iter=optimizer_info["checkpoint_after_iter"],
            checkpoint_load_iter=optimizer_info["checkpoint_load_iter"],
            weight_decay=optimizer_info["weight_decay"],
            momentum=optimizer_info["momentum"],
            drop_last=optimizer_info["drop_last"],
            pin_memory=optimizer_info["pin_memory"],
            num_workers=optimizer_info["num_workers"],
            seed=optimizer_info["seed"],
        )

        weights_path = path.parent / meta_data["model_paths"]
        self._load_model_weights(weights_path)

        return self

    def _load_model_weights(self, weights_path: Union[str, Path]):
        """Load pretrained model weights

        Args:
            weights_path (Union[str, Path]): Path to model weights file.
                Type of file must be one of {".pyth", ".pth", ".onnx"}
        """
        weights_path = Path(weights_path)

        assert weights_path.is_file() and weights_path.suffix in {".pyth", ".pth", ".onnx"}, (
            f"weights_path ({str(weights_path)}) should be a .pth or .onnx file."
            "Pretrained weights can be downloaded using `self.download(...)`"
        )
        if weights_path.suffix == ".onnx":
            return self._load_onnx(weights_path.parent)

        logger.debug(f"Loading model weights from {str(weights_path)}")

        loaded_state_dict = torch.load(weights_path, map_location=torch.device(self.device))
        self.model.load_state_dict(loaded_state_dict, strict=False)

        return self

    @staticmethod
    def download(path: Union[str, Path]):
        """Download pretrained models. As this module

        Args:
            path (Union[str, Path], optional): Directory in which to store model weights. Defaults to None.
            model_names (Iterable[str], optional): iterable with model names to download.
                The iterable may contain {"xs", "s", "m", "l"}.
                Defaults to _MODEL_NAMES.
        """
        raise NotImplementedError(
            "No pretrained models available. Please train your own model using the `fit` function."
        )

    def reset(self):
        pass

    def fit(
        self,
        dataset: Dataset,
        val_dataset: Dataset = None,
        epochs: int = None,
        steps: int = None,
        *args,
        **kwargs,
    ):
        """Fit the model to a dataset

        Args:
            dataset (Dataset): Training dataset
            val_dataset (Dataset, optional): Validation dataset.
                If none is given, validation steps are skipped. Defaults to None.
            epochs (int, optional): Number of epochs. If none is supplied, self.iters will be used. Defaults to None.
            steps (int, optional): Number of training steps to conduct. If none, this is determined by epochs. Defaults to None.
        """

        train_dataloader = torch.utils.data.DataLoader(
            dataset,
            batch_size=self.batch_size,
            num_workers=self._num_workers,
            shuffle=True,
            pin_memory=self._pin_memory,
            drop_last=self._drop_last,
        )
        val_dataloader = (
            torch.utils.data.DataLoader(
                val_dataset,
                batch_size=self.batch_size,
                num_workers=self._num_workers,
                shuffle=False,
                pin_memory=self._pin_memory,
                drop_last=self._drop_last,
            )
            if val_dataset
            else None
        )

        optimisation_metric = "val/loss" if val_dataset else "train/loss"

        # Patch model optimizer
        assert self.optimizer in {
            "adam",
            "sgd",
        }, f"Invalid optimizer '{self.optimizer}'. Must be 'adam' or 'sgd'."
        if self.optimizer == "adam":
            Optimizer = partial(
                torch.optim.Adam,
                lr=self.lr,
                betas=(self._momentum, 0.999),
                weight_decay=self._weight_decay,
            )
        else:  # self.optimizer == "sgd":
            Optimizer = partial(
                torch.optim.SGD,
                lr=self.lr,
                momentum=self._momentum,
                weight_decay=self._weight_decay,
            )

        def configure_optimizers():
            # nonlocal Optimizer, optimisation_metric
            optimizer = Optimizer(self.model.parameters())
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, patience=10)
            return {
                "optimizer": optimizer,
                "lr_scheduler": scheduler,
                "monitor": optimisation_metric,
            }

        self._plmodel.configure_optimizers = configure_optimizers

        self.trainer = pl.Trainer(
            max_epochs=epochs or self.iters,
            gpus=1 if "cuda" in self.device else 0,
            callbacks=[
                pl.callbacks.ModelCheckpoint(
                    save_top_k=1,
                    verbose=True,
                    monitor=optimisation_metric,
                    mode="min",
                    prefix="",
                )
            ],
            logger=_experiment_logger(),
        )
        self.trainer.limit_train_batches = steps or self.trainer.limit_train_batches
        self.trainer.limit_val_batches = steps or self.trainer.limit_val_batches

        self.trainer.fit(self._plmodel, train_dataloader, val_dataloader)
        # self.model.to(self.device)

    def eval(self, dataset: Dataset, steps: int = None) -> Dict[str, Any]:
        """Evaluate the model on the dataset

        Args:
            dataset (Dataset): Dataset on which to evaluate model
            steps (int, optional): Number of validation batches to evaluate.
                If None, all batches are evaluated. Defaults to None.

        Returns:
            Dict[str, Any]: Evaluation statistics
        """
        test_dataloader = torch.utils.data.DataLoader(
            dataset,
            batch_size=self.batch_size,
            num_workers=self._num_workers,
            shuffle=False,
            pin_memory=self._pin_memory,
            drop_last=False,
        )

        if not hasattr(self, "trainer"):
            self.trainer = pl.Trainer(
                gpus=1 if "cuda" in self.device else 0,
                logger=_experiment_logger(),
            )
        self.trainer.limit_test_batches = steps or self.trainer.limit_test_batches
        results = self.trainer.test(self._plmodel, test_dataloader)
        results = {
            "accuracy": results[-1]["test/acc"],
            "loss": results[-1]["test/loss"],
        }
        return results

    def infer(self, x: Union[Timeseries, Vector, torch.Tensor]) -> Category:
        """Run inference on a data point, x

        Args:
            x (Union[Timeseries, Vector, torch.Tensor])): Either a single time instance (Vector) or a Timeseries.
                x can also be passed as a torch.Tensor.

        Returns:
            Category: Network output
        """

        if isinstance(x, Vector):
            x = torch.tensor(x.data)
            # assert len(data) == self._input_dims
            forward_mode = "step"
            x = x.unsqueeze(0)  # Add batch dim
        elif isinstance(x, Timeseries):
            x = torch.tensor(x.data).permute(1, 0)
            assert x.shape == (self._input_dims, self._sequence_len)
            forward_mode = "regular"
            x = x.unsqueeze(0)  # Add batch dim
        else:
            assert isinstance(x, torch.Tensor)
            if len(x.shape) == 1:
                assert x.shape == (self._input_dims,)
                forward_mode = "step"
                x = x.unsqueeze(0)  # Add batch dim
            elif len(x.shape) == 2:
                if x.shape == (self.batch_size, self._input_dims):
                    forward_mode = "step"
                else:
                    assert x.shape == (self._input_dims, self._sequence_len)
                    forward_mode = "regular"
                    x = x.unsqueeze(0)  # Add batch dim
            else:
                assert len(x.shape) == 3
                assert x.shape == (
                    self.batch_size,
                    self._input_dims,
                    self._sequence_len,
                )
                forward_mode = "regular"

        x = x.to(device=self.device, dtype=torch.float)

        if self._ort_session is not None and self._ort_state is not None and forward_mode == "step":
            inputs = {
                "input": x.cpu().detach().numpy(),
                **self._ort_state,
            }
            r, *next_state = self._ort_session.run(None, inputs)
            r = torch.tensor(r)
            self._ort_state = {k: v for k, v in zip(self._ort_state.keys(), next_state)}
        else:
            self.model.eval()
            r = (self.model.forward if forward_mode == "regular" else self.model.forward_step)(x)
        if isinstance(r, torch.Tensor):
            r = torch.nn.functional.softmax(r[0], dim=-1)
            result = Category(prediction=int(r.argmax(dim=0)), confidence=r)
        else:
            # In the "continual inference"-mode, the model needs to warm up by first seeing
            # "sequence_len" time-steps. Until then, it will output co.TensorPlaceholder values
            result = Category(
                prediction=-1,
                confidence=torch.zeros(self._num_classes, dtype=torch.float),
            )
        return result

    def optimize(self, do_constant_folding=False):
        """Optimize model execution.
        This is accomplished by saving to the ONNX format and loading the optimized model.

        Args:
            do_constant_folding (bool, optional): Whether to optimize constants. Defaults to False.
        """
        if getattr(self.model, "_ort_session", None):
            logger.info("Model is already optimized. Skipping redundant optimization")
            return

        path = Path(self.temp_path or os.getcwd()) / "weights"
        if not path.exists():
            self._save_onnx(path, do_constant_folding)
        self._load_onnx(path)

    @property
    def _example_input(self):
        return torch.randn(1, self._input_dims, self._sequence_len).to(device=self.device)

    @_example_input.setter
    def _example_input(self):
        raise ValueError("_example_input is set thorugh 'sequence_len' and 'input_dims' parameters")

    def _save_onnx(self, path: Union[str, Path], do_constant_folding=False, verbose=False):
        """Save model in the ONNX format

        Args:
            path (Union[str, Path]): Directory in which to save ONNX model
            do_constant_folding (bool, optional): Whether to optimize constants. Defaults to False.
        """
        assert (
            int(torch.__version__.split(".")[1]) >= 10
        ), "ONNX optimization of the Continual Transformer Encoder requires torch >= 1.10.0."

        assert (
            int(getattr(co, "__version__", "0.0.0").split(".")[0]) >= 1
        ), "ONNX optimization of the Continual Transformer Encoder requires continual-inference >= 1.0.0."

        path = Path(path)
        path.mkdir(exist_ok=True, parents=True)

        self.model.eval()
        self.model.to(device="cpu")

        # Prepare state
        state0 = None
        with torch.no_grad():
            for i in range(self._sequence_len):
                _, state0 = self.model._forward_step(self._example_input[:, :, i], state0)
        state0 = co.utils.flatten(state0)

        # Export to ONNX
        onnx_path = path / "cotransenc_weights.onnx"
        logger.info(f"Saving model to ONNX format at {str(onnx_path)}")
        onnx.export(
            self.model,
            (self._example_input[:, :, -1], *state0),
            onnx_path,
            input_names=["input"],
            output_names=["output"],
            do_constant_folding=do_constant_folding,
            verbose=verbose,
            opset_version=14,
        )

        # Save default state and name mappings for later use
        state_path = path.parent / "cotransenc_state.pickle"
        logger.info(f"Saving ONNX model states at {str(state_path)}")
        omodel = onnx.OnnxWrapper(self.model)
        state = {k: v.detach().numpy() for k, v in zip(omodel.state_input_names, state0)}
        with open(state_path, "wb") as f:
            pickle.dump(state, f)

    def _load_onnx(self, path: Union[str, Path]):
        """Loads ONNX model into an onnxruntime inference session.

        Args:
            path (Union[str, Path]): Path to ONNX model folder
        """
        assert (
            int(getattr(ort, "__version__", "0.0.0").split(".")[1]) >= 11
        ), "ONNX inference of the Continual Transformer Encoder requires onnxruntime >= 1.11.0."

        onnx_path = path / "cotransenc_weights.onnx"
        state_path = path.parent / "cotransenc_state.pickle"

        logger.info(f"Loading ONNX runtime inference session from {str(onnx_path)}")
        self._ort_session = ort.InferenceSession(str(onnx_path))

        logger.info(f"Loading ONNX state from {str(state_path)}")
        with open(state_path, "rb") as f:
            self._ort_state = pickle.load(f)


def _experiment_logger():
    return pl.loggers.TensorBoardLogger(save_dir=Path(os.getcwd()) / "logs", name="cotransenc")


def _accuracy(x, y):
    return torch.sum(x.argmax(dim=1) == y) / len(y)


class _LightningModuleWithCrossEntropy(pl.LightningModule):
    def __init__(self, module):
        pl.LightningModule.__init__(self)
        self.module = module

    def forward(self, *args, **kwargs):
        return self.module.forward(*args, **kwargs)

    def training_step(self, batch, batch_idx):
        x, y = batch
        z = self.module(x)
        loss = torch.nn.functional.cross_entropy(z, y)
        self.log("train/loss", loss)
        self.log("train/acc", _accuracy(z, y))
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        z = self.forward(x)
        loss = torch.nn.functional.cross_entropy(z, y)
        self.log("val/loss", loss)
        self.log("val/acc", _accuracy(z, y))
        return loss

    def test_step(self, batch, batch_idx):
        x, y = batch
        z = self.forward(x)
        loss = torch.nn.functional.cross_entropy(z, y)
        self.log("test/loss", loss)
        self.log("test/acc", _accuracy(z, y))
        return loss



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import numpy as np
import math
from typing import Callable, Tuple
from torchvision.transforms import Compose
from torchvision.transforms._transforms_video import (
    CenterCropVideo,
    NormalizeVideo,
    RandomCropVideo,
    RandomHorizontalFlipVideo,
    ToTensorVideo,
)

Transform = Callable[[torch.Tensor], torch.Tensor]


def standard_video_transforms(
    spatial_pixels: int = 224,
    horizontal_flip=True,
    mean=(0.45, 0.45, 0.45),
    std=(0.225, 0.225, 0.225)
) -> Tuple[Transform, Transform]:
    """Generate standard transforms for video recognition

    Args:
        spatial_pixels (int, optional ): Spatial size (i.e. height or width) to resize to. Defaults to 224.
        horizontal_flip (bool, optional): Whether horizontal flipping (p = 0.5) is used. Defaults to True.
        mean (tuple, optional): Mean RGB values used in standardization. Defaults to (0.45, 0.45, 0.45).
        std (tuple, optional): Std RGB values used in standardization. Defaults to (0.225, 0.225, 0.225).

    Returns:
        Tuple[Transform, Transform]: [description]
    """
    train_scale = (1 / 0.7 * 0.8, 1 / 0.7)
    scaled_pix_min = (spatial_pixels * train_scale[0]) // 2 * 2
    scaled_pix_max = (spatial_pixels * train_scale[1]) // 2 * 2
    train_transforms = Compose(
        [
            t for t in [
                ToTensorVideo(),
                RandomShortSideScaleJitterVideo(
                    min_size=scaled_pix_min, max_size=scaled_pix_max
                ),
                RandomCropVideo(spatial_pixels),
                RandomHorizontalFlipVideo() if horizontal_flip else None,
                NormalizeVideo(mean=mean, std=std),
            ] if t
        ]
    )
    eval_transforms = Compose(
        [
            ToTensorVideo(),
            RandomShortSideScaleJitterVideo(min_size=spatial_pixels, max_size=spatial_pixels),
            CenterCropVideo(spatial_pixels),
            NormalizeVideo(mean=mean, std=std),
        ]
    )
    return train_transforms, eval_transforms


class RandomShortSideScaleJitterVideo:
    def __init__(self, min_size: int, max_size: int, inverse_uniform_sampling=False):
        """
        Args:
            min_size (int): the minimal size to scale the frames.
            max_size (int): the maximal size to scale the frames.
            inverse_uniform_sampling (bool): if True, sample uniformly in
                [1 / max_scale, 1 / min_scale] and take a reciprocal to get the
                scale. If False, take a uniform sample from [min_scale, max_scale].
        """
        self.min_size = min_size
        self.max_size = max_size
        self.inverse_uniform_sampling = inverse_uniform_sampling

    def __call__(self, images: torch.Tensor) -> torch.Tensor:
        """
        Perform a spatial short scale jittering on the given images and
        corresponding boxes.
        Args:
            images (tensor): images to perform scale jitter. Dimension is
                `num frames` x `channel` x `height` x `width`.
        Returns:
            (tensor): the scaled images with dimension of
                `num frames` x `channel` x `new height` x `new width`.
        """
        if self.inverse_uniform_sampling:
            size = int(
                round(1.0 / np.random.uniform(1.0 / self.max_size, 1.0 / self.min_size))
            )
        else:
            size = int(round(np.random.uniform(self.min_size, self.max_size)))

        height = images.shape[2]
        width = images.shape[3]
        if (width <= height and width == size) or (height <= width and height == size):
            return images
        new_width = size
        new_height = size
        if width < height:
            new_height = int(math.floor((float(height) / width) * size))
        else:
            new_width = int(math.floor((float(width) / height) * size))

        return torch.nn.functional.interpolate(
            images, size=(new_height, new_width), mode="bilinear", align_corners=False,
        )



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from opendr.engine.datasets import DatasetIterator


class DummyTimeseriesDataset(DatasetIterator, torch.utils.data.Dataset):
    """
    Dumme dataset for time-series forecasting

    Input data: Sinusoidal data of different wavelength
    Targets: Last sum of cosines quantized into four buckets
        (positive and falling, positive and rising, negative and falling, negative and rising)
    """

    def __init__(self, num_sines=4, num_datapoints=64, base_offset=0, sequence_len=64):
        DatasetIterator.__init__(self)
        torch.utils.data.Dataset.__init__(self)

        time_steps = torch.stack(
            [
                torch.stack(
                    [
                        torch.arange(
                            0 + offset + base_offset,
                            i + offset + base_offset,
                            i / sequence_len,
                        )
                        for i in range(1, num_sines + 1)
                    ],
                    dim=1,
                )
                for offset in range(num_datapoints)
            ]
        ).permute(0, 2, 1)
        self._input_data = torch.sin(time_steps)
        assert self._input_data.shape == (num_datapoints, num_sines, sequence_len)

        cosines = torch.cos(time_steps[:, :-2]).sum(dim=-1)
        positive = cosines[:, -1] > 0
        upwards = cosines[:, -1] > cosines[:, -2]
        self._output_data = positive + 2 * upwards

    def __getitem__(self, idx):
        return self._input_data[idx], self._output_data[idx]

    def __len__(self):
        return len(self._output_data)



import math
from collections import OrderedDict

import continual as co
import torch
from continual import PaddingMode
from torch import nn

from .res import CoResStage, init_weights

from opendr.perception.activity_recognition.x3d.algorithm.operators import Swish
from .se import CoSe


def CoX3DTransform(
    dim_in: int,
    dim_out: int,
    temp_kernel_size: int,
    stride: int,
    dim_inner: int,
    num_groups: int,
    stride_1x1=False,
    inplace_relu=True,
    eps=1e-5,
    bn_mmt=0.1,
    dilation=1,
    norm_module=torch.nn.BatchNorm3d,
    se_ratio=0.0625,
    swish_inner=True,
    block_idx=0,
    temporal_window_size: int = 4,
    temporal_fill: PaddingMode = "zeros",
    se_scope="frame",  # "frame" or "clip"
):
    """
    Args:
        dim_in (int): the channel dimensions of the input.
        dim_out (int): the channel dimension of the output.
        temp_kernel_size (int): the temporal kernel sizes of the middle
            convolution in the bottleneck.
        stride (int): the stride of the bottleneck.
        dim_inner (int): the inner dimension of the block.
        num_groups (int): number of groups for the convolution. num_groups=1
            is for standard ResNet like networks, and num_groups>1 is for
            ResNeXt like networks.
        stride_1x1 (bool): if True, apply stride to 1x1 conv, otherwise
            apply stride to the 3x3 conv.
        inplace_relu (bool): if True, calculate the relu on the original
            input without allocating new memory.
        eps (float): epsilon for batch norm.
        bn_mmt (float): momentum for batch norm. Noted that BN momentum in
            PyTorch = 1 - BN momentum in Caffe2.
        dilation (int): size of dilation.
        norm_module (torch.nn.Module): torch.nn.Module for the normalization layer. The
            default is torch.nn.BatchNorm3d.
        se_ratio (float): if > 0, apply SE to the Tx3x3 conv, with the SE
            channel dimensionality being se_ratio times the Tx3x3 conv dim.
        swish_inner (bool): if True, apply swish to the Tx3x3 conv, otherwise
            apply ReLU to the Tx3x3 conv.
    """
    (str1x1, str3x3) = (stride, 1) if stride_1x1 else (1, stride)

    a = co.Conv3d(
        dim_in,
        dim_inner,
        kernel_size=(1, 1, 1),
        stride=(1, str1x1, str1x1),
        padding=(0, 0, 0),
        bias=False,
    )

    a_bn = co.forward_stepping(
        norm_module(num_features=dim_inner, eps=eps, momentum=bn_mmt)
    )

    a_relu = torch.nn.ReLU(inplace=inplace_relu)

    # Tx3x3, BN, ReLU.
    b = co.Conv3d(
        dim_inner,
        dim_inner,
        kernel_size=(temp_kernel_size, 3, 3),
        stride=(1, str3x3, str3x3),
        padding=(int(temp_kernel_size // 2), dilation, dilation),
        groups=num_groups,
        bias=False,
        dilation=(1, dilation, dilation),
        temporal_fill=temporal_fill,
    )

    b_bn = co.forward_stepping(
        norm_module(num_features=dim_inner, eps=eps, momentum=bn_mmt)
    )

    # Apply SE attention or not
    use_se = True if (block_idx + 1) % 2 else False
    if se_ratio > 0.0 and use_se:
        se = CoSe(
            temporal_window_size,
            dim_in=dim_inner,
            ratio=se_ratio,
            temporal_fill=temporal_fill,
            scope=se_scope,
        )

    b_relu = co.forward_stepping(
        Swish()  # nn.SELU is the same as Swish
        if swish_inner
        else nn.ReLU(inplace=inplace_relu)
    )

    # 1x1x1, BN.
    c = co.Conv3d(
        dim_inner,
        dim_out,
        kernel_size=(1, 1, 1),
        stride=(1, 1, 1),
        padding=(0, 0, 0),
        bias=False,
    )
    c_bn = co.forward_stepping(
        norm_module(num_features=dim_out, eps=eps, momentum=bn_mmt)
    )
    c_bn.transform_final_bn = True

    return co.Sequential(
        OrderedDict(
            [
                ("a", a),
                ("a_bn", a_bn),
                ("a_relu", a_relu),
                ("b", b),
                ("b_bn", b_bn),
                *([("se", se)] if use_se else []),
                ("b_relu", b_relu),
                ("c", c),
                ("c_bn", c_bn),
            ]
        )
    )


def CoX3DHead(
    dim_in: int,
    dim_inner: int,
    dim_out: int,
    num_classes: int,
    pool_size: int,
    dropout_rate=0.0,
    act_func="softmax",
    inplace_relu=True,
    eps=1e-5,
    bn_mmt=0.1,
    norm_module=torch.nn.BatchNorm3d,
    bn_lin5_on=False,
    temporal_window_size: int = 4,
    temporal_fill: PaddingMode = "zeros",
    no_pool=False,
):
    """
    Continual X3D head.
    This layer performs a fully-connected projection during training, when the
    input size is 1x1x1. It performs a convolutional projection during testing
    when the input size is larger than 1x1x1. If the inputs are from multiple
    different pathways, the inputs will be concatenated after pooling.
    """
    modules = []
    modules.append(
        (
            "conv_5",
            co.Conv3d(
                dim_in,
                dim_inner,
                kernel_size=(1, 1, 1),
                stride=(1, 1, 1),
                padding=(0, 0, 0),
                bias=False,
            ),
        )
    )
    modules.append(
        ("conv_5_bn", norm_module(num_features=dim_inner, eps=eps, momentum=bn_mmt))
    )
    modules.append(("conv_5_relu", torch.nn.ReLU(inplace_relu)))

    if no_pool:
        return co.Sequential(OrderedDict(modules))

    avg_pool = co.Sequential(
        co.Lambda(lambda x: x.mean(dim=(-1, -2))),
        co.AvgPool1d(temporal_window_size, stride=1, temporal_fill=temporal_fill),
    )

    modules.append(("avg_pool", avg_pool))
    modules.append(
        (
            "lin_5",
            co.Conv1d(
                dim_inner,
                dim_out,
                kernel_size=(1,),
                stride=(1,),
                padding=(0,),
                bias=False,
            ),
        )
    )
    if bn_lin5_on:
        modules.append(
            ("lin_5_bn", torch.nn.BatchNorm1d(num_features=dim_out, eps=eps, momentum=bn_mmt))
        )

    modules.append(("lin_5_relu", torch.nn.ReLU(inplace_relu)))

    if dropout_rate > 0.0:
        modules.append(("dropout", torch.nn.Dropout(dropout_rate)))

    # Perform FC in a fully convolutional manner. The FC layer will be
    # initialized with a different std comparing to convolutional layers.
    modules.append(
        ("projection", co.Linear(dim_out, num_classes, bias=True, channel_dim=1))
    )

    def view(x):
        return x.view(x.shape[0], -1)

    modules.append(("view", co.Lambda(view, forward_only_fn=view)))

    return co.Sequential(OrderedDict(modules))


def CoX3DStem(
    dim_in: int,
    dim_out: int,
    kernel: int,
    stride: int,
    padding: int,
    inplace_relu=True,
    eps=1e-5,
    bn_mmt=0.1,
    norm_module=torch.nn.BatchNorm3d,
    temporal_fill: PaddingMode = "zeros",
    *args,
    **kwargs,
):
    """
    X3D's 3D stem module.
    Performs a spatial followed by a depthwise temporal Convolution, BN, and Relu followed by a
    spatiotemporal pooling.

    Args:
        dim_in (int): the channel dimension of the input. Normally 3 is used
            for rgb input, and 2 or 3 is used for optical flow input.
        dim_out (int): the output dimension of the convolution in the stem
            layer.
        kernel (list): the kernel size of the convolution in the stem layer.
            temporal kernel size, height kernel size, width kernel size in
            order.
        stride (list): the stride size of the convolution in the stem layer.
            temporal kernel stride, height kernel size, width kernel size in
            order.
        padding (int): the padding size of the convolution in the stem
            layer, temporal padding size, height padding size, width
            padding size in order.
        inplace_relu (bool): calculate the relu on the original input
            without allocating new memory.
        eps (float): epsilon for batch norm.
        bn_mmt (float): momentum for batch norm. Noted that BN momentum in
            PyTorch = 1 - BN momentum in Caffe2.
        norm_module (torch.nn.Module): torch.nn.Module for the normalization layer. The
            default is torch.nn.BatchNorm3d.
    """
    conv_xy = co.Conv3d(
        dim_in,
        dim_out,
        kernel_size=(1, kernel[1], kernel[2]),
        stride=(1, stride[1], stride[2]),
        padding=(0, padding[1], padding[2]),
        bias=False,
    )

    conv = co.Conv3d(
        dim_out,
        dim_out,
        kernel_size=(kernel[0], 1, 1),
        stride=(stride[0], 1, 1),
        padding=(padding[0], 0, 0),
        bias=False,
        groups=dim_out,
        temporal_fill=temporal_fill,
    )

    bn = norm_module(num_features=dim_out, eps=eps, momentum=bn_mmt)

    relu = torch.nn.ReLU(inplace_relu)

    # Wrap in sequential to match weight specification
    return co.Sequential(
        OrderedDict(
            [
                (
                    "pathway0_stem",
                    co.Sequential(
                        OrderedDict(
                            [
                                ("conv_xy", conv_xy),
                                ("conv", conv),
                                ("bn", bn),
                                ("relu", relu),
                            ]
                        )
                    ),
                )
            ]
        )
    )


def CoX3D(
    dim_in: int,
    image_size: int,
    temporal_window_size: int,
    num_classes: int,
    x3d_conv1_dim: int,
    x3d_conv5_dim: int,
    x3d_num_groups: int,
    x3d_width_per_group: int,
    x3d_width_factor: float,
    x3d_depth_factor: float,
    x3d_bottleneck_factor: float,
    x3d_use_channelwise_3x3x3: bool,
    x3d_dropout_rate: float,
    x3d_head_activation: str,
    x3d_head_batchnorm: bool,
    x3d_fc_std_init: float,
    x3d_final_batchnorm_zero_init: bool,
    temporal_fill: PaddingMode = "zeros",
    se_scope="frame",
    headless=False,
) -> co.Sequential:
    """
    Continual X3D model,
    adapted from https://github.com/facebookresearch/SlowFast

    Christoph Feichtenhofer.
    "X3D: Expanding Architectures for Efficient Video Recognition."
    https://arxiv.org/abs/2004.04730
    """
    norm_module = torch.nn.BatchNorm3d
    exp_stage = 2.0
    dim_conv1 = x3d_conv1_dim

    num_groups = x3d_num_groups
    width_per_group = x3d_width_per_group
    dim_inner = num_groups * width_per_group

    w_mul = x3d_width_factor
    d_mul = x3d_depth_factor

    dim_res1 = _round_width(dim_conv1, w_mul)
    dim_res2 = dim_conv1
    dim_res3 = _round_width(dim_res2, exp_stage, divisor=8)
    dim_res4 = _round_width(dim_res3, exp_stage, divisor=8)
    dim_res5 = _round_width(dim_res4, exp_stage, divisor=8)

    block_basis = [
        # blocks, c, stride
        [1, dim_res2, 2],
        [2, dim_res3, 2],
        [5, dim_res4, 2],
        [3, dim_res5, 2],
    ]

    # Basis of temporal kernel sizes for each of the stage.
    temp_kernel = [
        [5],  # conv1 temporal kernels.
        [3],  # res2 temporal kernels.
        [3],  # res3 temporal kernels.
        [3],  # res4 temporal kernels.
        [3],  # res5 temporal kernels.
    ]

    modules = []

    s1 = CoX3DStem(
        dim_in=dim_in,
        dim_out=dim_res1,
        kernel=temp_kernel[0] + [3, 3],
        stride=[1, 2, 2],
        padding=[temp_kernel[0][0] // 2, 1, 1],
        norm_module=norm_module,
        stem_func_name="x3d_stem",
        temporal_window_size=temporal_window_size,
        temporal_fill=temporal_fill,
    )
    modules.append(("s1", s1))

    # blob_in = s1
    dim_in = dim_res1
    dim_out = dim_in
    for stage, block in enumerate(block_basis):
        dim_out = _round_width(block[1], w_mul)
        dim_inner = int(x3d_bottleneck_factor * dim_out)

        n_rep = _round_repeats(block[0], d_mul)
        prefix = "s{}".format(stage + 2)  # start w res2 to follow convention

        s = CoResStage(
            dim_in=dim_in,
            dim_out=dim_out,
            dim_inner=dim_inner,
            temp_kernel_sizes=temp_kernel[1],
            stride=block[2],
            num_blocks=n_rep,
            num_groups=dim_inner if x3d_use_channelwise_3x3x3 else num_groups,
            num_block_temp_kernel=n_rep,
            trans_func=CoX3DTransform,
            stride_1x1=False,
            norm_module=norm_module,
            dilation=1,
            drop_connect_rate=0.0,
            temporal_window_size=temporal_window_size,
            temporal_fill=temporal_fill,
            se_scope=se_scope,
        )
        dim_in = dim_out
        modules.append((prefix, s))

    spat_sz = int(math.ceil(image_size / 32.0))
    head = CoX3DHead(
        dim_in=dim_out,
        dim_inner=dim_inner,
        dim_out=x3d_conv5_dim,
        num_classes=num_classes,
        pool_size=(temporal_window_size, spat_sz, spat_sz),
        dropout_rate=x3d_dropout_rate,
        act_func=x3d_head_activation,
        bn_lin5_on=bool(x3d_head_batchnorm),
        temporal_window_size=temporal_window_size,
        temporal_fill=temporal_fill,
        no_pool=headless,
    )
    modules.append(("head", head))
    seq = co.Sequential(OrderedDict(modules))
    init_weights(seq, x3d_fc_std_init, bool(x3d_final_batchnorm_zero_init))
    return seq


def _round_width(width, multiplier, min_depth=8, divisor=8):
    """Round width of filters based on width multiplier."""
    if not multiplier:
        return width

    width *= multiplier
    min_depth = min_depth or divisor
    new_filters = max(min_depth, int(width + divisor / 2) // divisor * divisor)
    if new_filters < 0.9 * width:
        new_filters += divisor
    return int(new_filters)


def _round_repeats(repeats, multiplier):
    """Round number of layers based on depth multiplier."""
    multiplier = multiplier
    if not multiplier:
        return repeats
    return int(math.ceil(multiplier * repeats))



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import json
import random
import torch
import zipfile
import os
from opendr.engine.datasets import ExternalDataset, DatasetIterator
from pathlib import Path
from logging import getLogger
from tqdm.contrib.concurrent import process_map
from functools import partial
from typing import List, Optional, Tuple, Union
from joblib import Memory
from opendr.perception.activity_recognition.datasets.utils import decoder
from opendr.perception.activity_recognition.datasets.utils.transforms import (
    standard_video_transforms,
)
from opendr.engine.constants import OPENDR_SERVER_URL
from urllib.request import urlretrieve

try:
    import av
except ModuleNotFoundError:
    try:
        import torchvision
        assert len(torchvision.__version__) > 0
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Either pyav (`pip install av`) or torchvision "
            "must be installed for the Kinetics loader to work"
        )

logger = getLogger(__file__)

with open(Path(__file__).parent / "kinetics400_classes.csv", "r") as file:
    my_reader = csv.reader(file, delimiter=",")
    CLASSES = [row[1] for i, row in enumerate(my_reader) if i != 0]


class KineticsDataset(ExternalDataset, DatasetIterator, torch.utils.data.Dataset):
    """
    Kinetics dataset for Trimmed Activity Recognition

    It has multiple versions available which are identified by their number of classes (400, 600, 700)

    Splits can be downloaded from:
    https://deepmind.com/research/open-source/open-source-datasets/kinetics/

    Videos must be downloaded using a YouTube crawler, e.g.:
    https://github.com/LukasHedegaard/youtube-dataset-downloader
    """

    def __init__(
        self,
        path: str,
        frames_per_clip: int,
        step_between_clips=1,
        temporal_downsampling=1,
        split="train",
        video_transform=None,
        use_caching=False,
        decoder_backend="pyav",
        spatial_pixels=224
    ):
        """
        Kinetics dataset

        This dataset consider every video as a collection of video clips of fixed size, specified
        by ``frames_per_clip``, where the step in frames between each clip is given by
        ``step_between_clips``.

        Args:
            path (str): path directory of the Kinetics dataset folder, containing subfolders "data" and "splits"
            frames_per_clip (int): number of frames in a clip
            step_between_clips (int): number of frames between each clip. Defaults to 1.
            temporal_downsampling (int): rate of downsampling in time. Defaults to 1.
            split (str, optional): Which split to use (Options are ["train", "val", "test"]). Defaults to "train".
            video_transform (callable, optional): A function/transform that takes in a TxHxWxC video
                and returns a transformed version. If None, a standard video transform will be applied.
                Defaults to None.
            use_caching (bool): Cache long-running operations. Defaults to False.
            decoder_backend (str): Name of library to use for video decoding
                (Options are ["pyav", "torchvision"]). Defaults to "pyav".
        """
        ExternalDataset.__init__(self, path=str(path), dataset_type="kinetics")
        DatasetIterator.__init__(self)
        torch.utils.data.Dataset.__init__(self)

        self.root_path = Path(path) / "data"
        self.annotation_path = Path(path) / "splits"
        self.num_retries = 10
        assert split in [
            "train",
            "val",
            "validate",
            "test",
        ], "Split '{}' not supported for Kinetics".format(split)
        if split == "val":
            split = "validate"
        self.split = split
        self.frames_per_clip = frames_per_clip
        self.step_between_clips = step_between_clips
        self.target_fps = 30
        self.temporal_downsampling = temporal_downsampling
        assert decoder_backend in {"pyav", "torchvision"}
        self.decoder_backend = decoder_backend

        if video_transform:
            self.video_transform = video_transform
        else:
            train_transform, eval_transform = standard_video_transforms(spatial_pixels=spatial_pixels)
            self.video_transform = (
                train_transform if self.split == "train" else eval_transform
            )

        validate_splits = (
            Memory(Path(os.getcwd()) / ".cache", verbose=1).cache(_validate_splits)
            if use_caching
            else _validate_splits
        )

        (
            self.labels,
            self.file_paths,
            self.spatial_temporal_idx,
            self.video_meta,
            self.classes,
            _,  # num_not_found,
            self.video_inds,
        ) = validate_splits(self.root_path, self.annotation_path, split)
        if len(self.classes) not in {400, 600, 700}:
            logger.warning(
                f"Only found {len(self.classes)} classes for {split} set, but expected either 400, 600, or 700 for Kinetics."
            )

    def __getitem__(self, idx):
        """
        This method is used for loading the idx-th sample of a dataset along with its annotation.

        :param idx: the index of the sample to load
        :return: the idx-th sample and its annotation
        :rtype: Tuple of (Data, Target)

        Returns tuple of:
            video (Tensor[T, H, W, C]): the `T` video frames
            label (int): class of the video clip
        """
        video_container = None
        frames = None
        for _ in range(self.num_retries):
            try:
                video_container = _get_video_container(
                    self.file_paths[idx],
                    multi_thread_decode=False,
                    backend=self.decoder_backend,
                )
            except Exception as e:
                logger.info(
                    "Failed to load video from {} with error {}".format(
                        self.file_paths[idx], e
                    )
                )
            # Select a random video if the current video was not able to access.
            if video_container is None:
                idx = random.randint(0, len(self.file_paths) - 1)
                continue

            # Decode video. Meta info is used to perform selective decoding.
            temporal_clip_idx = {
                "train": -1,  # pick random
                "validate": 0,  # pick middle of clip
                "test": 0,  # pick middle of clip
            }[self.split]
            frames = decoder.decode(
                video_container,
                self.temporal_downsampling,
                self.frames_per_clip,
                clip_idx=temporal_clip_idx,
                num_clips=1,
                video_meta=self.video_meta[idx],
                target_fps=self.target_fps,
                backend=self.decoder_backend,
            )

            # If decoding failed (wrong format, video is too short, and etc),
            # select another video.
            if frames is None:
                idx = random.randint(0, len(self.file_paths) - 1)
            else:
                break

        if frames is None:
            raise RuntimeError(
                "Failed to fetch video after {} retries.".format(self.num_retries)
            )

        video = frames
        label = self.labels[idx]

        if self.video_transform is not None:
            video = self.video_transform(video)

        sample = (video, label)

        return sample

    def __len__(self):
        """
        This method returns the size of the dataset.

        :return: the size of the dataset
        :rtype: int
        """
        return len(self.file_paths)

    @staticmethod
    def download_mini(path: Union[str, Path]):
        """Download mini version of dataset: One video of each class in Kinetics400

        Args:
            path (Union[str, Path]): Directory in which to store dataset
        """
        path = Path(path)
        if path.exists():
            logger.info("Kinetics400 mini already exists. Skipping download.")
            return

        path.mkdir(parents=True, exist_ok=True)

        url = os.path.join(
            OPENDR_SERVER_URL,
            "perception",
            "activity_recognition",
            "datasets",
            "kinetics400mini.zip",
        )
        zip_path = str(Path(path) / "kinetics400mini.zip")
        unzip_path = str(Path(path))

        logger.info(f"Downloading Kinetics400 mini from {url}")
        urlretrieve(url=url, filename=zip_path)

        logger.info(f"Unzipping Kinetics400 mini to {(unzip_path)}")
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(unzip_path)
        os.remove(zip_path)

    @staticmethod
    def download_micro(path: Union[str, Path]):
        """Download micro version of dataset: One video of first three classes in Kinetics400

        Args:
            path (Union[str, Path]): Directory in which to store dataset
        """
        path = Path(path)
        if path.exists():
            logger.info("Kinetics3 already exists. Skipping download.")
            return

        path.mkdir(parents=True, exist_ok=True)

        url = os.path.join(
            OPENDR_SERVER_URL,
            "perception",
            "activity_recognition",
            "datasets",
            "kinetics3.zip",
        )
        zip_path = str(Path(path) / "kinetics3.zip")
        unzip_path = str(Path(path))

        logger.info(f"Downloading Kinetics3 from {url}")
        urlretrieve(url=url, filename=zip_path)

        logger.info(f"Unzipping Kinetics3 to {(unzip_path)}")
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(unzip_path)
        os.remove(zip_path)


def _make_path_name(
    key_and_annotations: Tuple[str, dict], root_path: Path, extention="mp4"
) -> Optional[Tuple[Path, Optional[str]]]:
    key, annotation = key_and_annotations

    def map_segments(segment_arr: List[float]) -> Tuple[str, str]:
        assert len(segment_arr) == 2
        return (
            str(int(segment_arr[0])).zfill(6),
            str(int(segment_arr[1])).zfill(6),
        )

    start, stop = map_segments(annotation["annotations"]["segment"])
    if "label" in annotation["annotations"] and annotation["annotations"]["label"]:
        label = annotation["annotations"]["label"]
        p = (
            root_path / annotation["subset"] / label / f"{key}_{start}_{stop}.{extention}"
        ).resolve()
    else:
        p = (
            root_path / annotation["subset"] / f"{key}_{start}_{stop}.{extention}"
        ).resolve()
        label = None

    return (p, label) if p.exists() else None


def _validate_splits(
    root: str, annotation_path: str, split: str,
):
    root_path = Path(root)
    assert root_path.is_dir()

    annotation_file = Path(annotation_path) / f"{split}.json"
    assert annotation_file.exists()

    with open(annotation_file, "r") as f:
        annotations = json.load(f)

    # Validate annotations file
    test_key = next(iter(annotations))
    test_value = annotations[test_key]
    assert type(test_value) == dict
    assert "subset" in test_value
    assert "annotations" in test_value
    assert "label" in test_value["annotations"]
    assert "segment" in test_value["annotations"]

    # Determine valid files
    maybe_results = process_map(
        partial(_make_path_name, root_path=root_path,),
        annotations.items(),
        chunksize=4,
        desc=f"Validating Kinetics {split}",
    )

    results = list(filter(bool, maybe_results))
    num_not_found = len(maybe_results) - len(results)
    logger.info(
        f"{len(results)} / {len(maybe_results)} ({100*len(results)/len(maybe_results):.1f}%) of videos were valid"
    )

    assert len(results) > 0
    file_paths, labels = zip(*results)
    file_paths, labels = list(file_paths), list(labels)

    classes = sorted(set(labels))  # type: ignore
    class_to_idx = {n: i for i, n in enumerate(classes)}
    labels = [class_to_idx[lbl] for lbl in labels]
    file_paths = [str(p) for p in file_paths]

    num_clips = 1

    _labels = []
    _file_paths = []
    _spatial_temporal_idx = []
    _video_meta = {}
    _video_idx = []
    for i in range(len(file_paths)):
        for j in range(num_clips):
            _labels.append(labels[i])
            _file_paths.append(file_paths[i])
            _spatial_temporal_idx.append(j)
            _video_meta[i * num_clips + j] = {}
            _video_idx.append(i)

    return (
        _labels,
        _file_paths,
        _spatial_temporal_idx,
        _video_meta,
        classes,
        num_not_found,
        _video_idx,
    )


def _get_video_container(path_to_vid, multi_thread_decode=False, backend="pyav"):
    """
    Given the path to the video, return the pyav video container.
    Args:
        path_to_vid (str): path to the video.
        multi_thread_decode (bool): if True, perform multi-thread decoding.
        backend (str): decoder backend, options include `pyav` and
            `torchvision`, default is `pyav`.
    Returns:
        container (container): video container.
    """
    if backend == "torchvision":
        with open(path_to_vid, "rb") as fp:
            container = fp.read()
        return container
    elif backend == "pyav":
        container = av.open(path_to_vid)
        if multi_thread_decode:
            # Enable multiple threads for decoding.
            container.streams.video[0].thread_type = "AUTO"
        return container
    else:
        raise NotImplementedError(f"Unknown backend {backend}")



from collections import OrderedDict

import continual as co
import torch
from torch import Tensor, nn
from torch.nn.modules.pooling import AdaptiveAvgPool2d
from opendr.perception.activity_recognition.x3d.algorithm.operators import Swish


def _round_width(width, multiplier, min_width=8, divisor=8):
    """
    Round width of filters based on width multiplier
    Args:
        width (int): the channel dimensions of the input.
        multiplier (float): the multiplication factor.
        min_width (int): the minimum width after multiplication.
        divisor (int): the new width should be dividable by divisor.
    """
    if not multiplier:
        return width

    width *= multiplier
    min_width = min_width or divisor
    width_out = max(min_width, int(width + divisor / 2) // divisor * divisor)
    if width_out < 0.9 * width:
        width_out += divisor
    return int(width_out)


class SE(torch.nn.Module):
    """Squeeze-and-Excitation (SE) block w/ Swish: AvgPool, FC, Swish, FC, Sigmoid."""

    def __init__(self, dim_in, ratio, relu_act=True):
        """
        Args:
            dim_in (int): the channel dimensions of the input.
            ratio (float): the channel reduction ratio for squeeze.
            relu_act (bool): whether to use ReLU activation instead
                of Swish (default).
        """
        super(SE, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool3d((1, 1, 1))
        dim_fc = _round_width(dim_in, ratio)
        self.fc1 = nn.Conv3d(dim_in, dim_fc, 1, bias=True)
        self.fc1_act = nn.ReLU() if relu_act else Swish()
        self.fc2 = nn.Conv3d(dim_fc, dim_in, 1, bias=True)
        self.fc2_sig = nn.Sigmoid()
        self.dim_in = dim_in
        self.ratio = ratio
        self.relu_act = relu_act

    def forward(self, x):
        x_in = x
        for module in self.children():
            x = module(x)
        return x_in * x


def CoSe(
    window_size: int,
    dim_in: int,
    ratio: float,
    relu_act: bool = True,
    scope="frame",
    temporal_fill="zeros",
):
    dim_fc = _round_width(dim_in, ratio)
    return co.Residual(
        co.Sequential(
            OrderedDict(
                [
                    (
                        "avg_pool",
                        co.AdaptiveAvgPool3d(
                            output_size=(1, 1, 1),
                            kernel_size={
                                "clip": window_size,
                                "frame": 1,
                            }[scope],
                            temporal_fill=temporal_fill,
                        ),
                    ),
                    (
                        "fc1",
                        co.Conv3d(dim_in, dim_fc, 1, bias=True),
                    ),
                    (
                        "fc1_act",
                        nn.ReLU()
                        if relu_act
                        else Swish(),  # nn.SELU is the same as Swish
                    ),
                    (
                        "fc2",
                        co.Conv3d(dim_fc, dim_in, 1, bias=True),
                    ),
                    (
                        "fc2_sig",
                        nn.Sigmoid(),
                    ),
                ]
            )
        ),
        reduce="mul",
    )


class CoSeAlt(co.CoModule, nn.Module):
    """Continual Squeeze-and-Excitation (SE) block w/ Swish: AvgPool, FC, Swish, FC, Sigmoid."""

    def __init__(
        self,
        window_size: int,
        dim_in: int,
        ratio: float,
        relu_act: bool = True,
        scope="frame",
        temporal_fill="zeros",
    ):
        """
        Args:
            window_size (int): the window size to average over
            dim_in (int): the channel dimensions of the input.
            ratio (float): the channel reduction ratio for squeeze.
            relu_act (bool): whether to use ReLU activation instead
                of Swish (default).
        """
        super(CoSeAlt, self).__init__()
        self.avg_pool = {
            "clip": lambda: co.AdaptiveAvgPool3d(
                output_size=(1, 1, 1),
                kernel_size=window_size,
                temporal_fill=temporal_fill,
            ),
            "frame": lambda: co.forward_stepping(AdaptiveAvgPool2d(output_size=(1, 1))),
        }[scope]()
        dim_fc = _round_width(dim_in, ratio)
        self.fc1 = co.forward_stepping(nn.Conv3d(dim_in, dim_fc, 1, bias=True))
        self.fc1_act = nn.ReLU() if relu_act else Swish()
        self.fc2 = co.forward_stepping(nn.Conv3d(dim_fc, dim_in, 1, bias=True))
        self.fc2_sig = nn.Sigmoid()

        self.window_size = window_size
        self.dim_in = dim_in
        self.ratio = ratio
        self.relu_act = relu_act
        self.scope = scope
        self.temporal_fill = temporal_fill

    def forward(self, x: Tensor) -> Tensor:
        x_in = x
        for module in self.children():
            x = module(x)
        return x_in * x

    def forward_step(self, x: Tensor) -> Tensor:
        x_in = x
        for module in self.children():
            if hasattr(module, "forward_step"):
                x = module.forward_step(x)
            else:
                x = module(x)
            if not isinstance(x, Tensor):
                return None
        return x_in * x

    def forward_steps(self, x: Tensor) -> Tensor:
        x_in = x
        for module in self.children():
            if hasattr(module, "forward_steps"):
                x = module.forward_steps(x)
            else:
                x = module(x)
            if not isinstance(x, Tensor):
                return None
        return x_in * x

    @property
    def delay(self):
        return self.avg_pool.delay

    def clean_state(self):
        self.avg_pool.clean_state()

    def build_from(
        module: SE, window_size: int, scope="frame", temporal_fill="zeros"
    ) -> "CoSeAlt":
        mod = CoSeAlt(
            window_size,
            module.dim_in,
            module.ratio,
            module.relu_act,
            scope,
            temporal_fill,
        )
        mod.load_state_dict(module.state_dict())
        return mod



""" Adapted from: https://github.com/facebookresearch/SlowFast
"""

import torch
import torch.nn as nn


class Nonlocal(nn.Module):
    """
    Builds Non-local Neural Networks as a generic family of building
    blocks for capturing long-range dependencies. Non-local Network
    computes the response at a position as a weighted sum of the
    features at all positions. This building block can be plugged into
    many computer vision architectures.
    More details in the paper: https://arxiv.org/pdf/1711.07971.pdf
    """

    def __init__(
        self,
        dim,
        dim_inner,
        pool_size=None,
        instantiation="softmax",
        zero_init_final_conv=False,
        zero_init_final_norm=True,
        norm_eps=1e-5,
        norm_momentum=0.1,
        norm_module=nn.BatchNorm3d,
    ):
        """
        Args:
            dim (int): number of dimension for the input.
            dim_inner (int): number of dimension inside of the Non-local block.
            pool_size (list): the kernel size of spatial temporal pooling,
                temporal pool kernel size, spatial pool kernel size, spatial
                pool kernel size in order. By default pool_size is None,
                then there would be no pooling used.
            instantiation (string): supports two different instantiation method:
                "dot_product": normalizing correlation matrix with L2.
                "softmax": normalizing correlation matrix with Softmax.
            zero_init_final_conv (bool): If true, zero initializing the final
                convolution of the Non-local block.
            zero_init_final_norm (bool):
                If true, zero initializing the final batch norm of the Non-local
                block.
            norm_module (nn.Module): nn.Module for the normalization layer. The
                default is nn.BatchNorm3d.
        """
        super(Nonlocal, self).__init__()
        self.dim = dim
        self.dim_inner = dim_inner
        self.instantiation = instantiation
        self.norm_eps = norm_eps
        self.norm_momentum = norm_momentum

        # Three convolution heads: theta, phi, and g.
        self.conv_theta = nn.Conv3d(
            self.dim, self.dim_inner, kernel_size=1, stride=1, padding=0
        )
        self.conv_phi = nn.Conv3d(
            self.dim, self.dim_inner, kernel_size=1, stride=1, padding=0
        )
        self.conv_g = nn.Conv3d(
            self.dim, self.dim_inner, kernel_size=1, stride=1, padding=0
        )

        # Final convolution output.
        self.conv_out = nn.Conv3d(
            self.dim_inner, self.dim, kernel_size=1, stride=1, padding=0
        )
        # Zero initializing the final convolution output.
        self.conv_out.zero_init = zero_init_final_conv

        self.norm = norm_module(
            num_features=self.dim, eps=self.norm_eps, momentum=self.norm_momentum,
        )
        # Zero initializing the final bn.
        self.norm.transform_final_bn = zero_init_final_norm

        self.pool_size = pool_size
        use_pool = False if pool_size is None else any((size > 1 for size in pool_size))
        # Optional to add the spatial-temporal pooling.
        self.pool = (
            nn.MaxPool3d(
                kernel_size=self.pool_size,  # type:ignore
                stride=self.pool_size,
                padding=(0, 0, 0),
            )
            if use_pool
            else Dummy()
        )

    def forward(self, x: torch.Tensor):
        x_identity = x
        N, C, T, H, W = x.size()

        theta = self.conv_theta(x)

        # Perform temporal-spatial pooling to reduce the computation.
        x = self.pool(x)

        phi = self.conv_phi(x)
        g = self.conv_g(x)

        theta = theta.view(N, self.dim_inner, -1)
        phi = phi.view(N, self.dim_inner, -1)
        g = g.view(N, self.dim_inner, -1)

        # (N, C, TxHxW) * (N, C, TxHxW) => (N, TxHxW, TxHxW).
        theta_phi = torch.einsum("nct,ncp->ntp", (theta, phi))
        # For original Non-local paper, there are two main ways to normalize
        # the affinity tensor:
        #   1) Softmax normalization (norm on exp).
        #   2) dot_product normalization.
        if self.instantiation == "softmax":
            # Normalizing the affinity tensor theta_phi before softmax.
            theta_phi = theta_phi * (self.dim_inner ** -0.5)
            theta_phi = nn.functional.softmax(theta_phi, dim=2)
        elif self.instantiation == "dot_product":
            spatial_temporal_dim = theta_phi.shape[2]
            theta_phi = theta_phi / spatial_temporal_dim
        else:
            raise NotImplementedError("Unknown norm type {}".format(self.instantiation))

        # (N, TxHxW, TxHxW) * (N, C, TxHxW) => (N, C, TxHxW).
        theta_phi_g = torch.einsum("ntg,ncg->nct", (theta_phi, g))

        # (N, C, TxHxW) => (N, C, T, H, W).
        theta_phi_g = theta_phi_g.view(N, self.dim_inner, T, H, W)

        p = self.conv_out(theta_phi_g)
        p = self.norm(p)
        return x_identity + p


class Dummy(nn.Module):
    def forward(self, x):
        return x
